# Copyright 2018-2018 Berk Onat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from .TINKERCommon import PARSERNAME, PROGRAMNAME, PARSERTAG
from nomadcore.smart_parser.SmartParserDictionary import metaNameConverter, metaNameConverter_UnderscoreSpaceDash
from nomadcore.smart_parser.SmartParserDictionary import MetaInfoMap, FileInfoMap, MapDictionary
from nomadcore.smart_parser.SmartParserDictionary import getDict_MetaStrInDict, getList_MetaStrInDict, get_unitDict
import nomadcore.md_data_access.MDDataAccess as MDData

def get_fileListDict():
    """Loads dictionary for file namelists.

    Returns:
        the list of defaults file namelists
    """

    startpage = {
        'nameTranslate'   : metaNameConverter,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_file',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_input_output_files'],
        'fileInterface'   : ["tinkerread", "parmed", "mdanalysis", "mdtraj", "pymolfile"]
        }
    namelist = {
        'structure'     : FileInfoMap(startpage, fileFormat=['.tinkerxyz', '.txyz'], activeInfo=False,
                                      infoPurpose=['topology', 'inputcoordinates']),
        'input_coord'   : FileInfoMap(startpage, fileFormat=['.tinkerxyz', '.txyz'], activeInfo=False,
                                      infoPurpose=['inputcoordinates', 'inputvelocities']),
        'output_coord'  : FileInfoMap(startpage, fileFormat=['.tinkerdyn', '.dyn'], activeInfo=False,
                                      infoPurpose=['outputcoordinates', 'outputvelocities']),
        'trajectory'    : FileInfoMap(startpage, fileFormat=['.txyz', '.arc'], activeInfo=False,
                                      infoPurpose=['trajectory']),
        'parameter'     : FileInfoMap(startpage, fileFormat=['.tinkerprm'], activeInfo=False,
                                      infoPurpose=['forcefield']),
        'output_log'    : FileInfoMap(startpage),
        'control_input' : FileInfoMap(startpage),
        }
    return MapDictionary(namelist)

def get_nameListDict(deflist):
    """Loads control in data of TINKER.

    Args:
        deflist: name list definition (cntrl/parm).

    matchWith parameters:
        EOL = End of line
        PL  = Match with previous line
        PW  = Previous word
        NW  = Next word (until space, comma) (DEFAULT)
        UD  = until delimeter (can be any string/character)

        The text matched inside apostrophe or qoutation is extracted

    Returns:
        the list of namelists
    """
    startpage = {
        'nameTranslate'   : metaNameConverter_UnderscoreSpaceDash,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_control',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_control_parameters'],
        }

    mdcntrllist = {
        'RMS Grad'      : MetaInfoMap(startpage),
        'Molecular Dynamics Trajectory via'  : MetaInfoMap(startpage, matchWith='EOL',
            replaceTag='integrator',
            subFunc=lambda x: ' '.join(x.strip().split()) if x is not None else None),
        'Optimization'  : MetaInfoMap(startpage, matchWith='BOL',
            replaceTag='optimization algorithm',
            subFunc=lambda x: ' '.join(x.strip().split()) if x is not None else None),
        'Algorithm'     : MetaInfoMap(startpage),
        'Random Number Generator Initialized with SEED' : MetaInfoMap(startpage,
            replaceTag='Random Number Generator seed'),
        'Preconditioning' : MetaInfoMap(startpage),
        'Termin'      : MetaInfoMap(startpage, matchWith='PW', matchAlso='--',
            replaceTag='termination status'),
        'nation'     : MetaInfoMap(startpage, matchAlso='--', matchWith='EOL',
            replaceTag='termination reason',
            subFunc=lambda x: ' '.join(x.strip().replace(
                'due to', '').replace('at', '').split()) if x is not None else None),
        'Function Value' : MetaInfoMap(startpage),
        'RMS Gradient' : MetaInfoMap(startpage),
        'RMS Coordinate Gradient' : MetaInfoMap(startpage),
        'RMS Lattice Gradient' : MetaInfoMap(startpage),
        'Final Gradient Norm' : MetaInfoMap(startpage),
        'Coordinate Gradient Norm' : MetaInfoMap(startpage),
        'Lattice Gradient Norm' : MetaInfoMap(startpage),
        #''           : MetaInfoMap(startpage, defaultValue=None),
        }

    filecntrllist = {
        'key file'  : MetaInfoMap(startpage, defaultValue=None),
        'coordinate file list'  : MetaInfoMap(startpage, defaultValue=None),
        'topology file'  : MetaInfoMap(startpage, defaultValue=None),
        'force field file'  : MetaInfoMap(startpage, defaultValue=None),
        'initial configuration file'  : MetaInfoMap(startpage, defaultValue=None),
        'final configuration file'  : MetaInfoMap(startpage, defaultValue=None),
        'initial trajectory file'  : MetaInfoMap(startpage, defaultValue=None),
        'restart file'  : MetaInfoMap(startpage, defaultValue=None),
        'archive file'  : MetaInfoMap(startpage, defaultValue=None),
        }

    topocntrllist = {
        'atom'  : MetaInfoMap(startpage, defaultValue=None),
        'vdw'  : MetaInfoMap(startpage, defaultValue=None),
        'bond'  : MetaInfoMap(startpage, defaultValue=None),
        'angle'  : MetaInfoMap(startpage, defaultValue=None),
        'ureybond'  : MetaInfoMap(startpage, defaultValue=None),
        'charge'  : MetaInfoMap(startpage, defaultValue=None),
        }

    cntrllist = {
        'NSTEP'  : MetaInfoMap(startpage, defaultValue=None),
        'parameters'  : MetaInfoMap(startpage, defaultValue=None),
        'verbose'  : MetaInfoMap(startpage, defaultValue=None),
        'lights'  : MetaInfoMap(startpage, defaultValue=None),
        'randomseed'  : MetaInfoMap(startpage, defaultValue=None),
        'a-axis'  : MetaInfoMap(startpage, defaultValue=None),
        'b-axis'  : MetaInfoMap(startpage, defaultValue=None),
        'c-axis'  : MetaInfoMap(startpage, defaultValue=None),
        'alpha'  : MetaInfoMap(startpage, defaultValue=None),
        'beta'  : MetaInfoMap(startpage, defaultValue=None),
        'gamma'  : MetaInfoMap(startpage, defaultValue=None),
        'vdwtype'  : MetaInfoMap(startpage, defaultValue=None),
        'radiusrule'  : MetaInfoMap(startpage, defaultValue=None),
        'radiustype'  : MetaInfoMap(startpage, defaultValue=None),
        'radiussize'  : MetaInfoMap(startpage, defaultValue=None),
        'epsilonrule'  : MetaInfoMap(startpage, defaultValue=None),
        'dielectric'  : MetaInfoMap(startpage, defaultValue=None),
        'integrator'  : MetaInfoMap(startpage, defaultValue=None),
        'rattle'  : MetaInfoMap(startpage, defaultValue=None),
        'lambda'  : MetaInfoMap(startpage, defaultValue=None),
        'mutate'  : MetaInfoMap(startpage, defaultValue=None),
        'basin'  : MetaInfoMap(startpage, defaultValue=None),
        'saddlepoint'  : MetaInfoMap(startpage, defaultValue=None),
        'neighbor-list'  : MetaInfoMap(startpage, defaultValue=None),
        'vdw-cutoff'  : MetaInfoMap(startpage, defaultValue=None),
        'vdw-correction'  : MetaInfoMap(startpage, defaultValue=None),
        'ewald'  : MetaInfoMap(startpage, defaultValue=None),
        'ewald-cutoff'  : MetaInfoMap(startpage, defaultValue=None),
        'pme-grid'  : MetaInfoMap(startpage, defaultValue=None),
        'pme-order'  : MetaInfoMap(startpage, defaultValue=None),
        'polar-eps'  : MetaInfoMap(startpage, defaultValue=None),
        'enforce-chirality'  : MetaInfoMap(startpage, defaultValue=None),
        'printout'  : MetaInfoMap(startpage, defaultValue=None),
        'digits'  : MetaInfoMap(startpage, defaultValue=None),
        'spacegroup'  : MetaInfoMap(startpage, defaultValue=None),
        'vib-roots'  : MetaInfoMap(startpage, defaultValue=None),
        'group-inter'  : MetaInfoMap(startpage, defaultValue=None),
        'group'  : MetaInfoMap(startpage, defaultValue=None),
        'debug'  : MetaInfoMap(startpage, defaultValue=None),
        'chargeterm'  : MetaInfoMap(startpage, defaultValue=None),
        'archive'  : MetaInfoMap(startpage, defaultValue=None),
        'barostat'  : MetaInfoMap(startpage, defaultValue=None),
        'aniso-pressure'  : MetaInfoMap(startpage, defaultValue=None),
        'trial-distance'  : MetaInfoMap(startpage, defaultValue=None),
        'trial-distribution'  : MetaInfoMap(startpage, defaultValue=None),
        #'restrain-torsion'  : MetaInfoMap(startpage, defaultValue=None, activeInfo=False),
        #'restrain-distance'  : MetaInfoMap(startpage, defaultValue=None, activeInfo=False),
        'group-molecule'  : MetaInfoMap(startpage, defaultValue=None),
        'mpole-list'  : MetaInfoMap(startpage, defaultValue=None),
        'tau-temperature'  : MetaInfoMap(startpage, defaultValue=None),
        'tau-pressure'  : MetaInfoMap(startpage, defaultValue=None),
        }

    startpage.update({
        'metaNameTag'     : 'mdout',
        'activeSections'  : ['section_single_configuration_calculation']
        })
    mddatalist = {
        'MD Step'     : MetaInfoMap(startpage),
        'InputCoordStep'  : MetaInfoMap(startpage),
        'OutputCoordStep' : MetaInfoMap(startpage),
        'E Total'     : MetaInfoMap(startpage),
        'E Potential' : MetaInfoMap(startpage),
        'E Kinetic'   : MetaInfoMap(startpage),
        'Temp'        : MetaInfoMap(startpage),
        'Pres'        : MetaInfoMap(startpage),
        'STEP'        : MetaInfoMap(startpage),
        'TIME'        : MetaInfoMap(startpage),
        'VOLUME'      : MetaInfoMap(startpage),
        'VIRIAL'      : MetaInfoMap(startpage),
        'Dynamics Steps'    : MetaInfoMap(startpage, defaultValue=None, matchWith='PW'),
        'Simulation Time'   : MetaInfoMap(startpage, defaultValue=None),
        'Total Energy'      : MetaInfoMap(startpage, defaultValue=None),
        'Potential Energy'  : MetaInfoMap(startpage, defaultValue=None),
        'Kinetic Energy'    : MetaInfoMap(startpage, defaultValue=None),
        'Temperature'       : MetaInfoMap(startpage, defaultValue=None),
        'Pressure'          : MetaInfoMap(startpage, defaultValue=None),
        'Density'           : MetaInfoMap(startpage, defaultValue=None),
        'Current Time'      : MetaInfoMap(startpage, defaultValue=None),
        'Current Potential' : MetaInfoMap(startpage, defaultValue=None),
        'Current Kinetic'   : MetaInfoMap(startpage, defaultValue=None),
        'Lattice Lengths'   : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL',
            subFunc=lambda x: ','.join(x.strip().split()) if x is not None else None),
        'Lattice Angles'    : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL',
            subFunc=lambda x: ','.join(x.strip().split()) if x is not None else None),
        'Frame Number'      : MetaInfoMap(startpage, defaultValue=None),
        'Coordinate File'   : MetaInfoMap(startpage, defaultValue=None),
        'VM Iter'     : MetaInfoMap(startpage),
        'TN Iter'     : MetaInfoMap(startpage),
        'QN Iter'     : MetaInfoMap(startpage),
        'CG Iter'     : MetaInfoMap(startpage),
        'F Value'     : MetaInfoMap(startpage),
        'G RMS'       : MetaInfoMap(startpage),
        'RMS G'       : MetaInfoMap(startpage),
        'F Move'      : MetaInfoMap(startpage),
        'X Move'      : MetaInfoMap(startpage),
        'Angle'       : MetaInfoMap(startpage),
        'FG Call'     : MetaInfoMap(startpage),
        }

    startpage.update({
        'metaNameTag'     : 'parm',
        'activeSections'  : [PARSERTAG+'_mdin_method']
        })
    extralist = {
        'flags' : MetaInfoMap(startpage),
        'number_of_atoms' : MetaInfoMap(startpage),
        'minimization' : MetaInfoMap(startpage, activeInfo=True),
        'integrator' : MetaInfoMap(startpage, activeInfo=True),
        'thermostat' : MetaInfoMap(startpage, activeInfo=True),
        'barostat' : MetaInfoMap(startpage, activeInfo=True),
        'box_info' : MetaInfoMap(startpage),
        'unitcell_radius' : MetaInfoMap(startpage),
        'total_memory' : MetaInfoMap(startpage),
        'file_format' : MetaInfoMap(startpage),
        'file_version' : MetaInfoMap(startpage),
        'file_date' : MetaInfoMap(startpage),
        'file_time' : MetaInfoMap(startpage)
        }

    if deflist == 'mddata':
        namelist = mddatalist
    elif deflist == 'mdcntrl':
        namelist = mdcntrllist
    elif deflist == 'topocntrl':
        namelist = topocntrllist
    elif deflist == 'extra':
        namelist = extralist
    elif deflist == 'filecntrl':
        namelist = filecntrllist
    elif deflist == 'stepcontrol':
        namelist = stepcontrol
    else:
        namelist = cntrllist
    return MapDictionary(namelist)

def set_Dictionaries(self):
    self.unitDict = get_unitDict('si')
    self.fileDict = get_fileListDict()
    self.cntrlDict = get_nameListDict('cntrl')
    self.mdcntrlDict = get_nameListDict('mdcntrl')
    self.topocntrlDict = get_nameListDict('topocntrl')
    self.mdoutDict = get_nameListDict('mdout')
    self.filecntrlDict = get_nameListDict('filecntrl')
    #self.parmDict = get_nameListDict('parm')
    self.mddataDict = get_nameListDict('mddata')
    self.extraDict = get_nameListDict('extra')
    self.mdoutHeaderDict = {}
    #self.fileUnitDict = {}
    self.stepcontrolDict = {
            'logsteps'       : None,
            'nextlogsteps'   : None,
            'trajsteps'      : None,
            'velsteps'       : None,
            'forcesteps'     : None,
            'steps'          : None,
            'follow'         : None,
            'sectioncontrol' : None,
            'cntrlparmstep'  : [-1],
        }
    self.metaDicts = {
            'file'      : self.fileDict,
            'cntrl'     : self.cntrlDict,
            'mdcntrl'   : self.mdcntrlDict,
            'topocntrl'   : self.topocntrlDict,
            'filecntrl' : self.filecntrlDict,
            'mddata'    : self.mddataDict,
            'extra'     : self.extraDict,
            }
    self.sectionDict = {
            'section' : {
                "metaNameTag"      : ['input_output_files',
                                      'control_parameters'],
                "listTypStr"       : 'type_section',
                "repeatingSection" : False,
                "supraNames"       : ["section_run"]
                }
            }

def get_updateDictionary(self, defname):

    def value_convert_array(itemdict):
        val = itemdict["value"]
        if(val is not None and isinstance(val, str)):
            if("[" in val or "(" in val):
                return False, np.asarray(val), itemdict
        return False, val, itemdict

    def value_convert_array_norm(itemdict):
        val = []
        depdict = itemdict["depends"]
        for item in depdict["value"]:
            st=self.metaStorage.fetchAttrValue(item)
            if(st is not None and isinstance(st, str)):
                if("[" in st or "(" in st):
                    val.append(list(storedtext))
        if val:
            return False, np.linalg.norm(np.asarray(val)), itemdict
        else:
            return False, itemdict["value"], itemdict

    startpage = {
        'metaHeader'      : '',
        'metaNameTag'     : '',
        'activeMetaNames' : [],
        'activeSections'  : ['section_sampling_method','section_single_configuration_calculation',PARSERTAG+'_mdout']
        }

    # ---------------------------------------------------------------
    #   Definitions of meta data values for section_sampling_method
    # ---------------------------------------------------------------
    sampling = {
        #'ensemble_type' : MetaInfoMap(startpage, activeInfo=True,
        #    depends=[
        #        {'test' : [['DYNA', ' is not None'],
        #                   ['FIRSTT', ' in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\",'
        #                                  '\"langevin\"]'],
        #                   ['barostat', ' in [\"berendsen\",'
        #                                 '\"langevin\"]']],
        #         'assign' : 'NPT'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' in [\"langevin\"]'],
        #                   ['thermostat', ' not in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\"]'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'Langevin'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\"]'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'NVT'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' is None'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'NVE'},
        #        {'test' : [['minimization', ' is not None']],
        #         'assign' : 'minimization'},
        #        ],
        #    lookupdict=self.extraDict
        #    ),
        'sampling_method' : MetaInfoMap(startpage, activeInfo=True,
            depends=[
                {'test' : [['MINI', ' is not None']],
                 'assign' : 'geometry_optimization'},
                {'test' : [['DYNA', ' is not None']],
                 'assign' : 'molecular_dynamics'}
                ],
            lookupdict=self.cntrlDict
            ),
#            'settings_geometry_optimization' : MetaInfoMap(startpage),
#            'settings_metadynamics' : MetaInfoMap(startpage),
#            'settings_molecular_dynamics' : MetaInfoMap(startpage),
#            'settings_Monte_Carlo' : MetaInfoMap(startpage),
#        'geometry_optimization_energy_change' : MetaInfoMap(startpage,
#            depends={
#                '' : {'imin' : '1'},
#                },
#            lookupdict=self.cntrlDict
#            ),
#       'geometry_optimization_geometry_change' : MetaInfoMap(startpage),
        'geometry_optimization_method' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"SD\",\"sd\"]']],
                 'assign' : 'SD'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"ABNR\",\"abnr\"]']],
                 'assign' : 'Adopted Basis Newton-Raphson'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"NRAP\",\"nrap\"]']],
                 'assign' : 'Newton-Raphson'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"POWELL\",\"powell\"]']],
                 'assign' : 'Powell'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"POWE\",\"powe\"]']],
                 'assign' : 'Powell'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"TNPACK\",\"tnpack\"]']],
                 'assign' : 'Truncated Newton Method'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' not in [\"SD\",\"sd\",'
                                             '\"ABNR\",\"abnr\",'
                                             '\"NRAP\",\"nrap\",'
                                             '\"POWE\",\"powe\",'
                                             '\"POWELL\",\"powell\",'
                                             '\"TNPACK\",\"tnpack\"]']],
                 'assign' : 'CG'},
                ],
            lookupdict=self.cntrlDict,
            activeInfo=True,
            ),
#       'geometry_optimization_threshold_force' : MetaInfoMap(startpage),
        #'x_tinker_barostat_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['minimization', ' is None'],
        #                   ['barostat', '== \"berendsen\"']],
        #         'assign' : 'Berendsen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['barostat', '== \"langevin\"']],
        #         'assign' : 'Nose-Hoover Langevin'}
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_tinker_barostat_target_pressure' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['TARGET PRESSURE', ' is not None']],
        #         'value' : 'TARGET PRESSURE'}
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='atmosphere',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_tinker_barostat_tau' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['LANGEVIN OSCILLATION PERIOD', ' is not None']],
        #         'value' : 'LANGEVIN OSCILLATION PERIOD'},
        #        {'test' : [['BERENDSEN RELAXATION TIME', ' is not None']],
        #         'value' : 'BERENDSEN RELAXATION TIME'}
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='akmatime',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_tinker_integrator_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['integrator', ' is not None']],
        #         'value' : 'integrator'},
        #        {'test' : [['thermostat', ' == \"langevin\"']],
        #         'assign' : 'Langevin'}
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_integrator']
        #    ),
        'x_tinker_integrator_dt' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['DYNA', ' is not None'],
                           ['TIME STEP', ' is not None']],
                 'value' : 'TIME STEP'},
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='nano-seconds',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        'x_tinker_number_of_steps_requested' : MetaInfoMap(startpage,
            depends=[{'value' : 'NSTEP'}],
            lookupdict=self.cntrlDict,
            valtype='int',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        #'x_tinker_thermostat_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"tcouple\"']],
        #         'assign' : 'Berendsen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"rescale\"']],
        #         'assign' : 'Velocity Rescaling'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"reassign\"']],
        #         'assign' : 'Velocity Reassigning'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"andersen\"']],
        #         'assign' : 'Andersen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"langevin\"']],
        #         'assign' : 'Nose-Hoover Langevin'},
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_thermostat']
        #    ),
        'x_tinker_thermostat_target_temperature' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['DYNA', ' is not None'],
                           ['FINALT', ' is not None']],
                 'value' : 'FINALT'},
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='Kelvin',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            ),
        #'x_tinker_thermostat_tau' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['VELOCITY RESCALE FREQ', ' is not None']],
        #         'value' : 'VELOCITY RESCALE FREQ'},
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['VELOCITY REASSIGNMENT FREQ', ' is not None']],
        #         'value' : 'VELOCITY REASSIGNMENT FREQ'},
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['LOWE-ANDERSEN RATE', ' is not None']],
        #         'value' : 'LOWE-ANDERSEN RATE'},
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='pico-second',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_thermostat']
        #    ),
        'x_tinker_periodicity_type' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['CRYSTAL', ' is not None']],
                 'assign' : 'periodic boundaries'},
                {'test' : [['CRYSTAL',' is None']],
                 'assign' : 'no periodic boundaries'},
                ],
            lookupdict=self.cntrlDict,
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            )
        }

    # ------------------------------------------------------------
    #   Definitions for section_single_configuration_calculation
    # ------------------------------------------------------------
    singleconfcalc = {
        #'atom_forces_type' : MetaInfoMap(startpage,
        #    depends=[{'assign' : 'Force Field'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_correction_entropy' : MetaInfoMap(startpage),
        'energy_current' : MetaInfoMap(startpage),
        #'energy_electrostatic' : MetaInfoMap(startpage,
        #    depends=[{'value' : 'E '}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    activeInfo=True,
        #    lookupdict=self.mddataDict
        #    ),
        'energy_free_per_atom' : MetaInfoMap(startpage),
        'energy_free' : MetaInfoMap(startpage),
        #'energy_method_current' : MetaInfoMap(startpage,
        #    depends=[{'assign' : 'Force Field'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_T0_per_atom' : MetaInfoMap(startpage),
        'energy_total_T0_per_atom' : MetaInfoMap(startpage),
        'energy_total_T0' : MetaInfoMap(startpage,
            depends=[{'value' : 'E Potential'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kcal/mol',
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'energy_total' : MetaInfoMap(startpage,
            depends=[{'value' : 'E Total'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kcal/mol',
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'hessian_matrix' : MetaInfoMap(startpage),
        'single_configuration_calculation_converged' : MetaInfoMap(startpage),
        'single_configuration_calculation_to_system_ref' : MetaInfoMap(startpage),
        'single_configuration_to_calculation_method_ref' : MetaInfoMap(startpage),
        'time_calculation' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_cpu1_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_cpu1_start' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_date_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_date_start' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_wall_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_wall_start' : MetaInfoMap(startpage),
        'stress_tensor_kind' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['MINI', ' is not None']],
                 'assign' : 'geometry_optimization'},
                {'test' : [['DYNA', ' is not None']],
                 'assign' : 'molecular_dynamics'}
                ],
            activeInfo=True,
            lookupdict=self.extraDict
            ),
        'stress_tensor_value' : MetaInfoMap(startpage)
        }

    # section_single_energy_van_der_Waals of section_single_configuration_calculation
    singlevdw = {
        'energy_van_der_Waals_value' : MetaInfoMap(startpage,
            depends=[{'value' : 'E Vdw'}],
            lookupdict=self.mddataDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='kcal/mol',
            #autoSections=True,
            activeSections=['section_energy_van_der_Waals']
            ),
        }

    # Info for section_restricted_uri
    restrictions = {
        'number_of_restricted_uri_files' : MetaInfoMap(startpage,
            value=int(1),
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_files' : MetaInfoMap(startpage,
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri'],
            subfunction={
                'function' : self.parameter_file_name,
                'supportDict' : self.cntrlDict,
                },
            ),
        'restricted_uri_license' : MetaInfoMap(startpage,
            value='TINKER License',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_restriction' : MetaInfoMap(startpage,
            value='any access',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_reason' : MetaInfoMap(startpage,
            value='propriety license',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_issue_authority' : MetaInfoMap(startpage,
            value='TINKER',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        }

    # ------------------------------------------
    #   Definitions for section_frame_sequence
    # ------------------------------------------
    frameseq = {
        #'frame_sequence_conserved_quantity_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TS'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_conserved_quantity_stats' : MetaInfoMap(startpage),
        #'frame_sequence_conserved_quantity' : MetaInfoMap(startpage,
        #    depends=[{'store' : ''}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kcal/mol',
        #    lookupdict=self.mddataDict
        #    ),
        'frame_sequence_continuation_kind' : MetaInfoMap(startpage),
        'frame_sequence_external_url' : MetaInfoMap(startpage),
        'frame_sequence_kinetic_energy_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'MD Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_kinetic_energy_stats' : MetaInfoMap(startpage),
        'frame_sequence_kinetic_energy' : MetaInfoMap(startpage,
            depends=[{'store' : 'E Kinetic'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kcal/mol',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_local_frames_ref' : MetaInfoMap(startpage),
        'frame_sequence_potential_energy_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'MD Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_potential_energy_stats' : MetaInfoMap(startpage),
        'frame_sequence_potential_energy' : MetaInfoMap(startpage,
            depends=[{'store' : 'E Potential'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kcal/mol',
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_pressure_average_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_pressure_average' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'PRSI'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='atmosphere',
        #    lookupdict=self.mddataDict
        #    ),
        'frame_sequence_pressure_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'MD Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_pressure_stats' : MetaInfoMap(startpage),
        'frame_sequence_pressure' : MetaInfoMap(startpage,
            depends=[{'store' : 'Pres'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='bar',
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_temperature_average_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TS'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_temperature_average' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TEMPAVG'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='Kelvin',
        #    lookupdict=self.mddataDict
        #    ),
        'frame_sequence_temperature_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'MD Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_temperature_stats' : MetaInfoMap(startpage),
        'frame_sequence_temperature' : MetaInfoMap(startpage,
            depends=[{'store' : 'Temp'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='Kelvin',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_time' : MetaInfoMap(startpage,
            depends=[{'store' : 'TIME'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='pico-second',
            # Calculates in AKMA units including inputs but prints in picoseconds.
            lookupdict=self.mddataDict
            ),
        'x_tinker_frame_sequence_density_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Dynamics Steps'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_tinker_frame_sequence_density' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'Density'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='1.0/Angstrom**3',
            lookupdict=self.mddataDict
            ),
        'x_tinker_frame_sequence_volume_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'MD Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_tinker_frame_sequence_volume' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'volume'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='Angstrom**3',
            lookupdict=self.mddataDict
            ),
        'x_tinker_frame_sequence_total_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'MD Step'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_tinker_frame_sequence_total_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E Total'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kcal/mol',
            lookupdict=self.mddataDict
            ),
        'x_tinker_frame_sequence_total_kinetic_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_tinker_frame_sequence_total_kinetic_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E Kinetic'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kcal/mol',
            lookupdict=self.mddataDict
            ),
        #'x_tinker_frame_sequence_virial_kinetic_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_virial_kinetic_energy_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'VIRKE'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit=self.unitDict['kcal/mol'],
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_boundary_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_boundary' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'BOUNDARY'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kcal/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_bond_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_bond_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'E_Bonds'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_angle_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_angle_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'E_Angles'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_proper_dihedral_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_proper_dihedral_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'E_Dihedral'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_improper_dihedral_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_improper_dihedral_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'E_Improper'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_cross_dihedral_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_cross_dihedral_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'E_Crossdihedral'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_vdw_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_vdw_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'E_Vdw'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_electrostatic_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_electrostatic_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'E_Non-bonded'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_covalent_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_tinker_frame_sequence_covalent_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'E_Covalent'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_to_sampling_ref' : MetaInfoMap(startpage),
        'geometry_optimization_converged' : MetaInfoMap(startpage,
            depends=[
                {'test' : [["MINI", " is not None"],
                           ['Minimization exiting with', ' is not None']],
                 'assign' : True},
                ],
            activeInfo=True,
            lookupdict=self.cntrlDict
            ),
        #'previous_sequence_ref' : MetaInfoMap(startpage)
        }

    frameseqend = {
        #'number_of_conserved_quantity_evaluations_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'frame_sequence_conserved_quantity_frames'
        #            )
        #        )
        #    ),
        'number_of_frames_in_sequence' : MetaInfoMap(startpage, activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_potential_energy_frames'
                    )
                )
            ),
        'number_of_kinetic_energies_in_sequence' : MetaInfoMap(startpage, activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_kinetic_energy_frames'
                    )
                )
            ),
        'number_of_potential_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_potential_energy_frames'
                    )
                )
            ),
        'number_of_pressure_evaluations_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_pressure_frames'
                    )
                )
            ),
        'number_of_temperatures_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_temperature_frames'
                    )
                )
            ),
        'x_tinker_number_of_density_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_tinker_frame_sequence_density_frames'
                    )
                )
            ),
        #'x_tinker_number_of_boundary_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_boundary_frames'
        #            )
        #        )
        #    ),
        #'x_tinker_number_of_bond_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_bond_energy_frames'
        #            )
        #        )
        #    ),
        #'x_tinker_number_of_angle_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_angle_energy_frames'
        #            )
        #        )
        #    ),
        #'x_tinker_number_of_electrostatic_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_electrostatic_energy_frames'
        #            )
        #        )
        #    ),
        #'x_tinker_number_of_vdw_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_vdw_energy_frames'
        #            )
        #        )
        #    ),
        #'x_tinker_number_of_proper_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_proper_dihedral_energy_frames'
        #            )
        #        )
        #    ),
        #'x_tinker_number_of_improper_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_improper_dihedral_energy_frames'
        #            )
        #        )
        #    ),
        #'x_tinker_number_of_cross_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_cross_dihedral_energy_frames'
        #            )
        #        )
        #    ),
        #'x_tinker_number_of_covalent_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_tinker_frame_sequence_covalent_energy_frames'
        #            )
        #        )
        #    ),
        #'previous_sequence_ref' : MetaInfoMap(startpage)
        }

    # ----------------------------------
    #   Definitions for section_system
    # ----------------------------------
    # section_system
    sec_system = {
        #'topology_ref' : MetaInfoMap(startpage),
        'atom_velocities' : MetaInfoMap(startpage),
        'configuration_raw_gid' : MetaInfoMap(startpage),
        'local_rotations' : MetaInfoMap(startpage),
        'number_of_atoms' : MetaInfoMap(startpage,
            depends=[{'value' : 'number_of_atoms'}],
            valtype='int',
            lookupdict=self.extraDict
            ),
        'number_of_sites' : MetaInfoMap(startpage),
        'number_of_symmetry_operations' : MetaInfoMap(startpage),
        'reduced_symmetry_matrices' : MetaInfoMap(startpage),
        'reduced_symmetry_translations' : MetaInfoMap(startpage),
        'SC_matrix' : MetaInfoMap(startpage),
        'spacegroup_3D_choice' : MetaInfoMap(startpage),
        'spacegroup_3D_hall' : MetaInfoMap(startpage),
        'spacegroup_3D_international' : MetaInfoMap(startpage),
        'spacegroup_3D_number' : MetaInfoMap(startpage),
        'spacegroup_3D_origin_shift' : MetaInfoMap(startpage),
        'spacegroup_3D_pointgroup' : MetaInfoMap(startpage),
        'spacegroup_3D_std_lattice' : MetaInfoMap(startpage),
        'spacegroup_3D_std_positions' : MetaInfoMap(startpage),
        'spacegroup_3D_std_types' : MetaInfoMap(startpage),
        'spacegroup_3D_trasformation_matrix' : MetaInfoMap(startpage),
        'spacegroup_3D_wyckoff' : MetaInfoMap(startpage),
        'symmorphic' : MetaInfoMap(startpage),
        'system_name' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_system'],
            subfunction={
                'function' : MDData.MDDataConverter.topology_system_name,
                'supportDict' : self.topoDict,
                },
            #functionbase=self
            ),
        'time_reversal_symmetry' : MetaInfoMap(startpage)
        }

    # section_configuration_core of section_system
    configuration_core = {
        'number_of_electrons' : MetaInfoMap(startpage,
            #value=0,
            #valtype='float',
            ),
        'atom_labels' : MetaInfoMap(startpage,
            #subfunction=self.system_atom_labels()
            ),
        'atom_positions' : MetaInfoMap(startpage,
            #subfunction=self.system_atom_positions()
            ),
        'configuration_periodic_dimensions' : MetaInfoMap(startpage,
            ),
        'embedded_system' : MetaInfoMap(startpage),
        'lattice_vectors' : MetaInfoMap(startpage,
            #subfunction=self.system_lattice_vectors()
            ),
        'simulation_cell' : MetaInfoMap(startpage,
            #subfunction=self.system_simulation_cell()
            )
        }

    # section_spacegroup_3D_operation of section_system
    spacegroup_op = {
            'spacegroup_3D_rotation' : MetaInfoMap(startpage),
            'spacegroup_3D_translation' : MetaInfoMap(startpage)
        }

    # section_system_to_system_refs of section_system
    sys_to_sys = {
            'system_to_system_kind' : MetaInfoMap(startpage),
            'system_to_system_ref' : MetaInfoMap(startpage)
        }

    # --------------------------------------------------------
    #   Definitions of meta data values for section_topology
    # --------------------------------------------------------
    # section_topology of section_run
    topology = {
        'atom_to_molecule' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_atom_to_mol,
                'supportDict' : self.topoDict,
                },
            #functionbase=self
            ),
        'molecule_to_molecule_type_map' : MetaInfoMap(startpage),
        'number_of_topology_atoms' : MetaInfoMap(startpage,
            depends=[{'value' : 'number_of_atoms'}],
            valtype='int',
            lookupdict=self.extraDict
            ),
        'number_of_topology_molecules' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_num_topo_mol,
                'supportDict' : self.topoDict,
                },
            valtype='int',
            ),
        'topology_force_field_name' : MetaInfoMap(startpage,
            depends=[{'value' : 'PARAMETER file'}],
            lookupdict=self.cntrlDict
            )
        }

    # section_atom_type of section_topology
    atom_type = {
        'atom_type_charge' : MetaInfoMap(startpage),
        'atom_type_mass' : MetaInfoMap(startpage),
        'atom_type_name' : MetaInfoMap(startpage)
        }

    # section_constraint of section_topology
    constraint = {
        'constraint_atoms' : MetaInfoMap(startpage),
        'constraint_kind' : MetaInfoMap(startpage),
        'constraint_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_constraint' : MetaInfoMap(startpage),
        'number_of_constraints' : MetaInfoMap(startpage)
        }

    # section_interaction of section_topology
    interaction = {
        'interaction_atoms' : MetaInfoMap(startpage),
        'interaction_kind' : MetaInfoMap(startpage),
        'interaction_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_interaction' : MetaInfoMap(startpage),
        'number_of_interactions' : MetaInfoMap(startpage)
        }

    # -------------------------------------------------------------
    #   Definitions of meta data values for section_molecule_type
    # -------------------------------------------------------------
    # section_molecule_type of section_topology
    mol_type = {
        'molecule_type_name' : MetaInfoMap(startpage),
        'number_of_atoms_in_molecule' : MetaInfoMap(startpage),
        'settings_atom_in_molecule' : MetaInfoMap(startpage)
        }

    # section_molecule_constraint of section_molecule_type
    mol_constraint = {
        'molecule_constraint_atoms' : MetaInfoMap(startpage),
        'molecule_constraint_kind' : MetaInfoMap(startpage),
        'molecule_constraint_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_molecule_constraint' : MetaInfoMap(startpage),
        'number_of_molecule_constraints' : MetaInfoMap(startpage)
        }

    # section_molecule_interaction of section_molecule_type
    mol_interaction = {
        'molecule_interaction_atoms' : MetaInfoMap(startpage),
        'molecule_interaction_kind' : MetaInfoMap(startpage),
        'molecule_interaction_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_molecule_interaction' : MetaInfoMap(startpage),
        'number_of_molecule_interactions' : MetaInfoMap(startpage)
        }

    # section_atom_in_molecule of section_molecule_type
    atom_in_mol = {
        'atom_in_molecule_charge' : MetaInfoMap(startpage),
        'atom_in_molecule_name' : MetaInfoMap(startpage),
        'atom_in_molecule_to_atom_type_ref' : MetaInfoMap(startpage)
        }


    if defname == 'system':
        dictionary = systemDict
    elif defname == 'topology':
        dictionary = topology
    elif defname == 'singleconfcalc':
        dictionary = singleconfcalc
    elif defname == 'restrictions':
        dictionary = restrictions
    elif defname == 'frameseq':
        dictionary = frameseq
    elif defname == 'frameseqend':
        dictionary = frameseqend
    elif defname == 'atom_type':
        dictionary = atom_type
    elif defname == 'molecule_type':
        dictionary = moltypeDict
    elif defname == 'interaction':
        dictionary = interaction
    elif defname == 'sampling':
        dictionary = sampling
    elif defname == 'singlevdw':
        dictionary = singlevdw
    elif defname == 'confcore':
        dictionary = configuration_core
    else:
        dictionary = singleconfcalclist
    return MapDictionary(dictionary)





# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import click

from nomad.cli.cli import cli


@cli.group(help='Commands related to the nomad source code.')
def dev():
    pass


@dev.command(help='Runs tests and linting of the nomad python source code. Useful before committing code.')
@click.option('--skip-tests', help='Do no tests, just do code checks.', is_flag=True)
@click.option('-x', '--exitfirst', help='Stop testing after first failed test case.', is_flag=True)
def qa(skip_tests: bool, exitfirst: bool):
    os.chdir(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    ret_code = 0
    if not skip_tests:
        click.echo('Run tests ...')
        ret_code += os.system('python -m pytest -sv%s tests' % ('x' if exitfirst else ''))
    click.echo('Run code style checks ...')
    ret_code += os.system('python -m pycodestyle --ignore=E501,E701,E731 nomad tests')
    click.echo('Run linter ...')
    ret_code += os.system('python -m pylint --load-plugins=pylint_mongoengine,nomad/metainfo/pylint_plugin nomad tests')
    click.echo('Run static type checks ...')
    ret_code += os.system('python -m mypy --ignore-missing-imports --follow-imports=silent --no-strict-optional nomad tests')

    sys.exit(ret_code)


@dev.command(help='Runs tests and linting of the nomad gui source code. Useful before committing code.')
def gui_qa():
    click.echo('Run gui code linting ...')
    os.chdir(os.path.abspath(os.path.join(os.path.dirname(__file__), '../../gui')))
    ret_code = 0
    ret_code += os.system('yarn run eslint \'src/**/*.js\'')
    sys.exit(ret_code)


@dev.command(help='Generates a JSON with all metainfo.')
def metainfo():
    import json
    print(json.dumps(metainfo_undecorated(), indent=2))


def metainfo_undecorated():
    from nomad.metainfo import Package, Environment

    # TODO the __init_metainfo__() should not be necessary and automatically performed
    # Also load and initialize the datamodel definitions
    import nomad.metainfo.metainfo
    import nomad.datamodel.datamodel
    import nomad.datamodel.dft
    import nomad.datamodel.ems
    import nomad.datamodel.optimade
    import nomad.datamodel.encyclopedia
    nomad.metainfo.metainfo.m_package.__init_metainfo__()
    nomad.datamodel.datamodel.m_package.__init_metainfo__()
    nomad.datamodel.dft.m_package.__init_metainfo__()  # pylint: disable=no-member
    nomad.datamodel.ems.m_package.__init_metainfo__()  # pylint: disable=no-member
    nomad.datamodel.optimade.m_package.__init_metainfo__()  # pylint: disable=no-member
    nomad.datamodel.encyclopedia.m_package.__init_metainfo__()

    # Ensure all parser metainfo is loaded
    from nomad.parsing.parsers import parsers
    for parser in parsers:
        _ = parser.metainfo_env

    export = Environment()
    for package in Package.registry.values():
        export.m_add_sub_section(Environment.packages, package)

    return export.m_to_dict(with_meta=True)


@dev.command(help='Generates a JSON with all search quantities.')
def search_quantities():
    from nomad import search
    # Due to this import, the parsing module will register all code_names based on parser
    # implementations.
    from nomad.parsing.parsers import parser_dict  # pylint: disable=unused-import
    import json

    def to_dict(search_quantity):
        result = {
            'name': search_quantity.qualified_name,
            'description': search_quantity.description,
            'many': search_quantity.many,
        }

        if search_quantity.statistic_fixed_size is not None:
            result['statistic_size'] = search_quantity.statistic_fixed_size
        if search_quantity.statistic_values is not None:
            result['statistic_values'] = search_quantity.statistic_values

        return result

    export = {
        search_quantity.qualified_name: to_dict(search_quantity)
        for search_quantity in search.search_quantities.values()
    }
    print(json.dumps(export, indent=2))


@dev.command(help='Generates source-code for the new metainfo from .json files of the old.')
@click.argument('path', nargs=-1)
def legacy_metainfo(path):
    from nomad.metainfo.legacy import convert, generate_metainfo_code

    if len(path) == 0:
        path = [
            'abinit.nomadmetainfo.json',
            'aptfim.nomadmetainfo.json',
            'atk.nomadmetainfo.json',
            'band.nomadmetainfo.json',
            'bigdft.nomadmetainfo.json',
            'castep.nomadmetainfo.json',
            'cp2k.nomadmetainfo.json',
            'cpmd.nomadmetainfo.json',
            'crystal.nomadmetainfo.json',
            'dl_poly.nomadmetainfo.json',
            'dmol3.nomadmetainfo.json',
            'eels.nomadmetainfo.json',
            'elastic.nomadmetainfo.json',
            'elk.nomadmetainfo.json',
            'exciting.nomadmetainfo.json',
            'fhi_aims.nomadmetainfo.json',
            'fleur.nomadmetainfo.json',
            'gamess.nomadmetainfo.json',
            'gaussian.nomadmetainfo.json',
            'gpaw.nomadmetainfo.json',
            'gulp.nomadmetainfo.json',
            'lib_atoms.nomadmetainfo.json',
            'molcas.nomadmetainfo.json',
            'mpes.nomadmetainfo.json',
            'nwchem.nomadmetainfo.json',
            'octopus.nomadmetainfo.json',
            'onetep.nomadmetainfo.json',
            'orca.nomadmetainfo.json',
            'phonopy.nomadmetainfo.json',
            'photoemission.nomadmetainfo.json',
            'qbox.nomadmetainfo.json',
            'quantum_espresso.nomadmetainfo.json',
            'siesta.nomadmetainfo.json',
            'turbomole.nomadmetainfo.json',
            'vasp.nomadmetainfo.json',
            'wien2k.nomadmetainfo.json',
            'dft.nomadmetainfo.json',
            'ems.nomadmetainfo.json']

    for element in path:
        env = convert(element)
        generate_metainfo_code(env)


@dev.command(help='Generates a JSON file that compiles all the parser metadata from each parser project.')
def parser_metadata():
    import inspect
    import re
    import json
    import yaml

    from nomad.parsing import LegacyParser, FairdiParser
    from nomad.parsing.parsers import parser_dict

    parsers_metadata = {}
    for parser in parser_dict.values():
        if isinstance(parser, LegacyParser):
            parser_class = parser.parser_class
        elif isinstance(parser, FairdiParser):
            parser_class = parser.__class__
        else:
            continue
        parser_code_file = inspect.getfile(parser_class)

        path_match = re.match(r'(.*/dependencies/parsers/[^/]+)/.*', parser_code_file)
        if path_match:
            parser_metadata_file = os.path.join(path_match.group(1), 'metadata.yaml')
            if os.path.exists(parser_metadata_file):
                with open(parser_metadata_file) as f:
                    parser_metadata = yaml.load(f, Loader=yaml.FullLoader)
                parsers_metadata[parser.code_name] = parser_metadata

    parsers_metadata = {
        key: parsers_metadata[key]
        for _, key in sorted([(key.lower(), key) for key in parsers_metadata], key=lambda x: x[0])}

    print(json.dumps(parsers_metadata, indent=2))


@dev.command(help='Generate toolkit tutorial metadata from anaytics submodules.')
def toolkit_metadata():
    import requests
    import re
    import json
    modules = requests.get(
        'https://gitlab.mpcdf.mpg.de/api/v4/projects/3161/repository/files/.gitmodules/raw?ref=master').text

    tutorials = []
    lines = modules.split('\n')
    for line in lines:
        match = re.match(r'\s*url = (.*)$', line)
        if match:
            url = match.group(1).replace('.git', '') + '/-/raw/master/metainfo.json'
            response = requests.get(url)
            if response.status_code != 200:
                print('Could not get metadata for %s' % match.group(1), file=sys.stderr)
                continue
            try:
                tutorials.append(response.json())
            except Exception:
                print('Could not get metadata for %s. Project is probably not public.' % match.group(1), file=sys.stderr)

    print(json.dumps(dict(tutorials=tutorials), indent=2))


@dev.command(help=(
    'Updates parser`s README files by combining a general template with  '
    'a parser`s metadata YAML file.'))
def update_parsers_readmes():
    from glob import glob
    import re
    import yaml
    print('\nPWD: ', os.getcwd())
    print('WARNING: to be run from project`s root dir\n')

    # based on
    # https://www.kite.com/python/answers/how-to-update-and-replace-text-in-a-file-in-python

    # filenames
    local_fn = 'README.md'
    generic_fn = './README.parsers.md'
    parser_path = './dependencies/parsers/'

    for num, ddir in enumerate(sorted(glob(parser_path + '*/')), start=1):
        parser_dirname = ddir.split('/')[-2]
        print('{} Working on {}' .format(num, parser_dirname))

        # Open general template
        with open(generic_fn, 'r') as generic:  # read only
            body = generic.read()

        # Open local YAML metadata file
        local_mdata = ddir + 'metadata.yaml'
        with open(local_mdata, 'r') as mdata_f:
            mdata = yaml.load(mdata_f, Loader=yaml.FullLoader)

        # Find & Replace Parser`s metadata on its README file
        local_readme = ddir + local_fn
        with open(local_readme, 'w') as local:
            for key in mdata.keys():
                ignore = ['codeLabelStyle', 'parserDirName']
                if key in ignore:
                    continue

                print('\tReplacing', key)
                find = r'\$' + key + r'\$'
                replace = mdata[key]

                if key == 'parserSpecific':
                    if mdata[key] != '':
                        replace = r'---\n## Parser Specific\n' + replace
                # print('\t', find, ' -> ', replace)
                body = re.sub(find, replace, body)

            # Extra: to replace the codeName (there's no YAML key for this)
            key = 'codeName'
            print('\tReplacing', key)
            find = r'\$' + key + r'\$'
            replace = parser_dirname
            body = re.sub(find, replace, body)

            # Extra: to replace the comment at the top of the gereral template
            find = (
                r'\*\*\*Note:\*\* This is a general README file for NOMAD parsers, '
                r'consult the README of specific parser projects for more detailed '
                r'information!\*\n\n')
            replace = ''
            print('\tReplacing the top comment')
            # print('\t', find, ' -> ', replace)
            body = re.sub(find, replace, body)

            # save file
            local.seek(0)  # go to the top
            local.write(body)
            local.truncate()


@dev.command(help='Updates parser`s metadata YAML')
def update_parsers_metadata():
    import re
    from glob import glob
    print('\nPWD: ', os.getcwd())
    print('WARNING: to be run from project`s root dir\n')

    # Based on
    # https://www.kite.com/python/answers/how-to-update-and-replace-text-in-a-file-in-python

    # dictionary to change YAML keys to camelCase
    new_keys = {
        'code-label': 'codeLabel',
        'code-label-style': 'codeLabelStyle',
        'code-url': 'codeUrl',
        'parser-dir-name': 'parserDirName',
        'parser-git-url': 'parserGitUrl',
        'parser-specific': 'parserSpecific',
        'table-of-files': 'tableOfFiles'}

    num = 0
    parser_path = './dependencies/parsers/'
    for ddir in sorted(glob(parser_path + '*/')):
        num += 1
        parser_dirname = ddir.split('/')[-2]
        print('{} Working on {}' .format(num, parser_dirname))

        local_mdata = ddir + 'metadata.yaml'
        if os.path.isfile(local_mdata) is False:
            print('\tWARNING: old readme file absent: {}'.format(local_mdata))
            continue

        # Replace key names in metadata.yaml
        # we open it as plain text: YAML module is unnecesary
        with open(local_mdata, 'r+') as local:
            body = local.read()

            for key in new_keys.keys():
                # Add  ':' to ensure complete sentence
                find, replace = key + ':', new_keys[key] + ':'

                # print('\t {:15} \t -> {} ' .format(find, replace))
                body = re.sub(find, replace, body)

            # save file
            local.seek(0)  # go to the top
            local.write(body)
            local.truncate()


@dev.command(help='Creates a Javascript source file containing the required unit conversion factors.')
@click.pass_context
def units(ctx):
    import re
    import json
    from nomad.units import ureg

    # Mapping from unit name to dimension
    unit_map = {
        # Time
        "second": {
            "dimension": "time",
            "label": "Second",
            "abbreviation": "s",
        },
        "atomic_unit_of_time": {
            "dimension": "time",
            "label": "Atomic unit of time",
            "abbreviation": "atomic_unit_of_time",
        },
        # Length
        "meter": {
            "dimension": "length",
            "label": "Meter",
            "abbreviation": "m",
        },
        "bohr": {
            "dimension": "length",
            "label": "Bohr",
            "abbreviation": "bohr",
        },
        "angstrom": {
            "dimension": "length",
            "label": "Ångstrom",
            "abbreviation": "Å",
        },
        # Mass
        "kilogram": {
            "dimension": "mass",
            "label": "Kilogram",
            "abbreviation": "kg",
        },
        "electron_mass": {
            "dimension": "mass",
            "label": "Electron mass",
            "abbreviation": "mₑ",
        },
        "unified_atomic_mass_unit": {
            "dimension": "mass",
            "label": "Unified atomic mass unit",
            "abbreviation": "u",
        },
        # Current
        "ampere": {
            "dimension": "current",
            "label": "Ampere",
            "abbreviation": "A",
        },
        "atomic_unit_of_current": {
            "dimension": "current",
            "label": "Atomic unit of current",
            "abbreviation": "atomic_unit_of_current",
        },
        # Substance
        "mole": {
            "dimension": "substance",
            "label": "Mole",
            "abbreviation": "mole",
        },
        # Luminosity
        "candela": {
            "dimension": "luminosity",
            "label": "Candela",
            "abbreviation": "cd",
        },
        # Temperature
        "kelvin": {
            "dimension": "temperature",
            "label": "Kelvin",
            "abbreviation": "K",
        },
        "celsius": {
            "dimension": "temperature",
            "label": "Celsius",
            "abbreviation": "°C",
        },
        "fahrenheit": {
            "dimension": "temperature",
            "label": "Fahrenheit",
            "abbreviation": "°F",
        },
        "atomic_unit_of_temperature": {
            "dimension": "temperature",
            "label": "Atomic unit of temperature",
            "abbreviation": "atomic_unit_of_temperature",
        },
        # Force
        "newton": {
            "dimension": "force",
            "label": "Newton",
            "abbreviation": "N",
        },
        "atomic_unit_of_force": {
            "dimension": "force",
            "label": "Atomic unit of force",
            "abbreviation": "atomic_unit_of_force",
        },
        # Pressure
        "pascal": {
            "dimension": "pressure",
            "label": "Pascal",
            "abbreviation": "Pa"
        },
        # Energy
        "joule": {
            "dimension": "energy",
            "label": "Joule",
            "abbreviation": "J",
        },
        "electron_volt": {
            "dimension": "energy",
            "label": "Electron volt",
            "abbreviation": "eV",
        },
        "hartree": {
            "dimension": "energy",
            "label": "Hartree",
            "abbreviation": "Ha",
        },
        # Power
        "watt": {
            "dimension": "power",
            "label": "Watt",
            "abbreviation": "W",
        },
        # Frequency
        "hertz": {
            "dimension": "frequency",
            "label": "Hertz",
            "abbreviation": "Hz",
        },
        # Electric potential
        "volt": {
            "dimension": "electric_potential",
            "label": "Volt",
            "abbreviation": "V",
        },
        # Capacitance
        "farad": {
            "dimension": "capacitance",
            "label": "Farad",
            "abbreviation": "F",
        },
        # Charge
        "coulomb": {
            "dimension": "charge",
            "label": "Coulomb",
            "abbreviation": "C",
        },
        "elementary_charge": {
            "dimension": "charge",
            "label": "Elementary charge",
            "abbreviation": "e",
        },
        # Magnetic field
        "tesla": {
            "dimension": "magnetic_field",
            "label": "Tesla",
            "abbreviation": "T",
        },
        # Magnetic flux
        "weber": {
            "dimension": "magnetic_flux",
            "label": "Weber",
            "abbreviation": "Wb",
        },
        # Inductance
        "henry": {
            "dimension": "inductance",
            "label": "Henry",
            "abbreviation": "H",
        },
        # dimensionless
        "dimensionless": {
            "dimension": "dimensionless",
            "label": "Dimensionless",
            "abbreviation": "",
        },
    }

    # Units that are supported
    unit_table = {
        # Base units
        "time": {
            "units": [
                "second",
                "atomic_unit_of_time",
            ],
            "multipliers": {},
        },
        "length": {
            "units": [
                "meter",
                "bohr",
                "angstrom",
            ],
            "multipliers": {},
        },
        "mass": {
            "units": [
                "kilogram",
                "electron_mass",
                "unified_atomic_mass_unit",
            ],
            "multipliers": {},
        },
        "current": {
            "units": [
                "ampere",
                "atomic_unit_of_current",
            ],
            "multipliers": {},
        },
        "substance": {
            "units": [
                "mole",
            ],
            "multipliers": {},
        },
        "luminosity": {
            "units": [
                "candela",
            ],
            "multipliers": {},
        },
        "temperature": {
            "units": [
                "kelvin",
                "celsius",
                "fahrenheit",
                "atomic_unit_of_temperature",
            ],
            "multipliers": {},
            "constants": {},
        },
        # Derived units
        "force": {
            "units": [
                "newton",
                "atomic_unit_of_force",
            ],
            "multipliers": {},
        },
        "pressure": {
            "units": [
                "pascal",
            ],
            "multipliers": {},
        },
        "energy": {
            "units": [
                "joule",
                "electron_volt",
                "hartree",
            ],
            "multipliers": {},
        },
        "power": {
            "units": [
                "watt",
            ],
            "multipliers": {},
        },
        "frequency": {
            "units": [
                "hertz",
            ],
            "multipliers": {},
        },
        "electric_potential": {
            "units": [
                "volt",
            ],
            "multipliers": {},
        },
        "capacitance": {
            "units": [
                "farad",
            ],
            "multipliers": {},
        },
        "charge": {
            "units": [
                "coulomb",
                "elementary_charge",
            ],
            "multipliers": {},
        },
        "magnetic_field": {
            "units": [
                "tesla",
            ],
            "multipliers": {},
        },
        "magnetic_flux": {
            "units": [
                "weber",
            ],
            "multipliers": {},
        },
        "inductance": {
            "dimension": "inductance",
            "units": [
                "henry",
            ],
            "multipliers": {},
        },
        "dimensionless": {
            "dimension": "dimensionless",
            "units": [
                "dimensionless",
            ],
            "multipliers": {},
        },
    }

    # Unit systems
    unit_systems = {
        "SI": {
            "label": "SI",
            "description": "International System of Units (SI)",
            "units": {
                "time": "second",
                "length": "meter",
                "mass": "kilogram",
                "current": "ampere",
                "substance": "mole",
                "luminosity": "candela",
                "temperature": "kelvin",
                "force": "newton",
                "pressure": "pascal",
                "energy": "joule",
                "power": "watt",
                "frequency": "hertz",
                "electric_potential": "volt",
                "charge": "coulomb",
            },
        },
        "AU": {
            "label": "Atomic units",
            "description": "Hartree atomic units",
            "units": {
                "time": "atomic_unit_of_time",
                "length": "bohr",
                "mass": "electron_mass",
                "current": "atomic_unit_of_current",
                "temperature": "atomic_unit_of_temperature",
                "force": "atomic_unit_of_force",
                "energy": "hartree",
            }
        }
    }

    # Precompute conversion factors and possible shifts
    for value in unit_table.values():
        units = value["units"]
        for i_unit in units:
            for j_unit in units:
                # Create dictionaries if not present
                multipliers = value["multipliers"]
                if i_unit not in multipliers:
                    multipliers[i_unit] = {}
                if j_unit not in multipliers[i_unit]:
                    multipliers[i_unit][j_unit] = {}

                # Check if there is a constant shift: y = ax + b -> y(0) = b.
                # Uses delta units for temperatures since only they are
                # multiplicative.
                y_0 = 0 * getattr(ureg, "delta_" + i_unit, getattr(ureg, i_unit))
                b = y_0.to(getattr(ureg, j_unit)).magnitude

                # Solving the multiplication factor with:
                # y(1) = a + b -> a = y(1) - b
                # Uses delta units for temperatures since only they are
                # multiplicative. Causes minor numerical accuracy issues in the
                # factor due to floating point precision
                y_1 = 1 * getattr(ureg, "delta_" + i_unit, getattr(ureg, i_unit))
                a = y_1.to(getattr(ureg, "delta_" + j_unit, getattr(ureg, j_unit))).magnitude
                multipliers[i_unit][j_unit] = a

                # Create dictionaries if not present
                if b != 0:
                    constants = value["constants"]
                    if i_unit not in constants:
                        constants[i_unit] = {}
                    if j_unit not in constants[i_unit]:
                        constants[i_unit][j_unit] = {}
                    constants[i_unit][j_unit] = b

    # Check that all defined units are present in the conversion table
    for unit, info in unit_map.items():
        dimension = info["dimension"]
        unit_list = unit_table[dimension]["units"]
        assert unit in unit_list, "Could not find '{}' in the unit table under the dimension '{}'.".format(unit, dimension)

    # Check unit system correctness
    for system in unit_systems.values():
        for dimension, unit in system["units"].items():
            info = unit_map[unit]
            assert dimension == info["dimension"]

    # Go through the metainfo and check that all units are defined
    all_metainfo = metainfo_undecorated()
    units = set()
    packages = all_metainfo["packages"]
    for package in packages:
        sections = package["section_definitions"]
        for section in sections:
            quantities = section.get("quantities", [])
            for quantity in quantities:
                unit = quantity.get("unit")
                if unit is not None:
                    parts = unit.split()
                    for part in parts:
                        is_operator = part in {"/", "**", "*"}
                        is_number = True
                        try:
                            int(part)
                        except Exception:
                            is_number = False
                        if not is_operator and not is_number:
                            units.add(part)
    for unit in units:
        assert unit in unit_map, "The unit '{}' is not defined in the unit definitions.".format(unit)

    # Print unit conversion table and unit systems as a Javascript source file
    output = "// Generated by NOMAD CLI. Do not edit manually.\n"
    output += "export const unitMap = "
    json_string = json.dumps(unit_map, indent=2)
    json_string = re.sub(r'(?<!: )"(\S*?)":', '\\1:', json_string)
    json_string = json_string.replace("\"", "'")
    output += json_string
    output += "\nexport const conversionMap = "
    json_string = json.dumps(unit_table, indent=2)
    json_string = re.sub(r'(?<!: )"(\S*?)":', '\\1:', json_string)
    json_string = json_string.replace("\"", "'")
    output += json_string
    output += "\nexport const unitSystems = "
    json_string = json.dumps(unit_systems, indent=2)
    json_string = re.sub(r'(?<!: )"(\S*?)":', '\\1:', json_string)
    json_string = json_string.replace("\"", "'")
    output += json_string
    print(output)



# pylint: disable=undefined-variable
from .exceptions import *  # noqa: F403
from .references import *  # noqa: F403
from .structures import *  # noqa: F403


__all__ = exceptions.__all__ + references.__all__ + structures.__all__  # noqa: F405



import os
import numpy as np
from ase import Atoms

from nomad.parsing.text_parser import Quantity, UnstructuredTextFileParser


class ElasticProperties:
    def __init__(self, mainfile):
        self._mainfile = mainfile
        self.maindir = os.path.dirname(mainfile)
        self._deform_dirs = None
        self._deform_dir_prefix = 'Dst'
        self._info = None

    @property
    def mainfile(self):
        return self._mainfile

    @mainfile.setter
    def mainfile(self, val):
        self._info = None
        self._deform_dirs = None
        self._mainfile = val

    @property
    def info(self):
        if self._info is None:

            quantities = [
                Quantity('order', r'\s*Order of elastic constants\s*=\s*([0-9]+)'),
                Quantity('calculation_method', r'\s*Method of calculation\s*=\s*([-a-zA-Z]+)\s*'),
                Quantity('code_name', r'\s*DFT code name\s*=\s*([-a-zA-Z]+)'),
                Quantity('space_group_number', r'\s*Space-group number\s*=\s*([0-9]+)'),
                Quantity('equilibrium_volume', r'\s*Volume of equilibrium unit cell\s*=\s*([0-9.]+)\s*'),
                Quantity('max_strain', r'\s*Maximum Lagrangian strain\s*=\s*([0-9.]+)'),
                Quantity('n_strains', r'\s*Number of distorted structures\s*=\s*([0-9]+)')
            ]

            parser = UnstructuredTextFileParser(self.mainfile, quantities)

            self._info = {}
            for key, val in parser.items():
                self._info[key] = val[0]

        return self._info

    @property
    def deformation_dirs(self):
        if self._deform_dirs is None:
            dirs = os.listdir(self.maindir)
            self._deform_dirs = [
                os.path.join(self.maindir, d) for d in dirs if d.startswith(self._deform_dir_prefix)]

        return self._deform_dirs

    def get_references_to_calculations(self):
        def output_file(dirname):
            code = self.info['code_name'].lower()
            if code == 'exciting':
                return os.path.join(dirname, 'INFO.OUT')
            elif code == 'wien':
                return os.path.join(dirname, '%s_Converged.scf' % os.path.basename(dirname))
            elif code == 'quantum':
                return os.path.join(dirname, '%s.out' % os.path.basename(dirname))
            else:
                return None

        references = []
        for deform_dir in self.deformation_dirs:
            sub_dirs = os.listdir(deform_dir)
            for sub_dir in sub_dirs:
                calc_dir = os.path.join(deform_dir, sub_dir)
                out_file = output_file(calc_dir)
                if out_file is not None and os.path.isfile(out_file):
                    references.append(out_file)

        return references

    def get_structure_info(self):
        path = os.path.join(self.maindir, 'sgroup.out')
        if not os.path.isfile(path):
            return

        def get_sym_pos(val):
            val = val.strip().replace('\n', '').split()
            sym = []
            pos = []
            for i in range(0, len(val), 4):
                sym.append(val[i + 3].strip())
                pos.append([float(val[j]) for j in range(i, i + 3)])
            sym_pos = dict(symbols=sym, positions=pos)
            return sym_pos

        quantities = [
            Quantity('cellpar', r'a\s*b\s*c\n([\d\.\s]+)\n\s*alpha\s*beta\s*gamma\n([\d\.\s]+)\n+'),
            Quantity('sym_pos', r'Atom positions:\n\n([\s\d\.A-Za-z]+)\n\n', str_operation=get_sym_pos)
        ]

        parser = UnstructuredTextFileParser(path, quantities)

        cellpar = list(parser['cellpar'])[0]
        sym_pos = list(parser['sym_pos'])[0]

        sym = list(sym_pos['symbols'])
        pos = list(sym_pos['positions'])

        structure = Atoms(cell=cellpar, scaled_positions=pos, symbols=sym, pbc=True)

        return sym, structure.get_positions(), list(structure.get_cell())

    def get_strain_energy(self):
        strains, energies = [], []

        for deform_dir in self.deformation_dirs:
            filenames = [d for d in os.listdir(deform_dir) if d.endswith('Energy.dat')]
            if not filenames:
                continue

            path = os.path.join(deform_dir, filenames[-1])
            data = np.loadtxt(path).T
            strains.append(data[0])
            energies.append(data[1])

        return strains, energies

    def get_strain_stress(self):
        strains = {'Lagrangian-stress': [], 'Physical-stress': []}
        stresses = {'Lagrangian-stress': [], 'Physical-stress': []}

        for deform_dir in self.deformation_dirs:
            filenames = [d for d in os.listdir(deform_dir) if d.endswith('stress.dat')]

            for filename in filenames:
                path = os.path.join(deform_dir, filename)
                if not os.path.isfile(path):
                    continue

                with open(path) as f:
                    lines = f.readlines()

                strain, stress = [], []
                for line in lines:
                    val = line.strip().split()
                    if not val[0].strip().replace('.', '').isdecimal():
                        continue

                    strain.append(float(val[0]))
                    stress.append([float(v) for v in val[1:7]])

                stype = filename.rstrip('.dat').split('_')[-1]
                strains[stype].append(strain)
                stresses[stype].append(stress)

        return strains, stresses

    def get_deformation_types(self):
        path = os.path.join(self.maindir, 'Distorted_Parameters')
        if not os.path.isfile(path):
            return

        quantities = [
            Quantity(
                'deformation', r'Lagrangian strain\s*=\s*\(([eta\s\d\.,]+)\)',
                str_operation=lambda x: x.replace(',', '').split(), dtype=str)]
        parser = UnstructuredTextFileParser(path, quantities)

        deformation_types = [list(d) for d in list(parser['deformation'])]

        return deformation_types

    def _get_fit(self, path_dir, file_ext):
        path_dir = os.path.join(self.maindir, path_dir)

        if not os.path.isdir(path_dir):
            return

        paths = [p for p in os.listdir(path_dir) if p.endswith(file_ext)]
        paths.sort()

        if not paths:
            return

        eta = {'2nd': [], '3rd': [], '4th': [], '5th': [], '6th': [], '7th': []}
        val = {'2nd': [], '3rd': [], '4th': [], '5th': [], '6th': [], '7th': []}

        def split_eta_val(val):
            val = val.strip().split()
            return [val[0::2], val[1::2]]

        quantities = [
            Quantity(
                order, r'%s order fit.\n([\d.\s\ne\-\+]+)\n' % order, str_operation=split_eta_val)
            for order in eta.keys()
        ]

        parser = UnstructuredTextFileParser(os.path.join(path_dir, paths[0]), quantities)

        for path in paths:
            parser.mainfile = os.path.join(path_dir, path)
            for key in eta.keys():
                if parser[key] is not None:
                    eta[key].append(list(parser[key][0][0]))
                    val[key].append(list(parser[key][0][1]))

        eta = {key: val for key, val in eta.items() if val}
        val = {key: val for key, val in val.items() if val}

        return eta, val

    def get_energy_fit(self):
        energy_fit = dict()

        for file_ext in ['d2E.dat', 'd3E.dat', 'ddE.dat']:
            result = self._get_fit('Energy-vs-Strain', file_ext)
            if result is None:
                continue

            energy_fit['d2e'] = result

        result = self._get_fit('Energy-vs-Strain', 'CVe.dat')
        if result is not None:
            energy_fit['cross-validation'] = result

        return energy_fit

    def get_stress_fit(self):
        stress_fit = dict()
        stress_fit['dtn'] = [[]] * 6
        stress_fit['cross-validation'] = [[]] * 6

        for strain_index in range(1, 7):
            result = self._get_fit('Stress-vs-Strain', '%d_dS.dat' % strain_index)
            if result is not None:
                stress_fit['dtn'][strain_index - 1] = result

            result = self._get_fit('Stress-vs-Strain', '%d_CVe.dat' % strain_index)
            if result is not None:
                stress_fit['cross-validation'][strain_index - 1] = result

        return stress_fit

    def get_input(self):
        paths = os.listdir(self.maindir)
        path = None
        order = self.info['order']
        for p in paths:
            if 'ElaStic_' in p and p.endswith('.in') and str(order) in p:
                path = p
                break

        if path is None:
            return

        calc_method = self.info['calculation_method']

        eta_ec = []
        fit_ec = []

        def _is_number(var):
            try:
                float(var)
                return True
            except Exception:
                return False

        path = os.path.join(self.maindir, path)

        with open(path) as f:
            while True:
                line = f.readline()
                if not line:
                    break

                if calc_method.lower() == 'energy':
                    _, eta, fit = line.strip().split()
                    eta_ec.append(float(eta))
                    fit_ec.append(int(fit))

                elif calc_method.lower() == 'stress':
                    val = line.strip().split()
                    if not _is_number(val[0]):
                        eta_ec.append([float(val[i + 1]) for i in range(6)])
                    else:
                        fit_ec.append([int(val[i]) for i in range(6)])

                else:
                    pass

        return eta_ec, fit_ec

    def get_elastic_constants_order2(self):
        path = os.path.join(self.maindir, 'ElaStic_2nd.out')

        def reshape(val):
            return np.array(val.split()).reshape((6, 6))

        quantities = [
            Quantity(
                'voigt', r'Symmetry[\s\S]+\n\s*\n([C\d\s\n\(\)\-\+\/\*]+)\n',
                str_operation=reshape, dtype=str),
            Quantity(
                'elastic_constant', r'Elastic constant[\s\S]+in GPa\s*:\s*\n\n([\-\d\.\s\n]+)\n',
                str_operation=reshape, dtype=float),
            Quantity(
                'compliance', r'Elastic compliance[\s\S]+in 1/GPa\s*:\s*\n\n([\-\d\.\s\n]+)\n',
                str_operation=reshape, dtype=float)]

        modulus_names = [
            'B_V', 'K_V', 'G_V', 'B_R', 'K_R', 'G_R', 'B_H', 'K_H', 'G_H', 'E_V',
            'nu_V', 'E_R', 'nu_R', 'E_H', 'nu_H', 'AVR']

        quantities += [
            Quantity(name, r'%s\s*=\s*([-+\d+\.]+)\s*' % name) for name in modulus_names]

        quantities += [
            Quantity(
                'eigenvalues', r'Eigenvalues of elastic constant \(stiffness\) matrix:\s*\n+([\-\d\.\n\s]+)\n')]

        parser = UnstructuredTextFileParser(path, quantities)

        matrices = dict(
            voigt=list(parser['voigt'])[0], elastic_constant=list(parser['elastic_constant'])[0],
            compliance=list(parser['compliance'])[0])

        moduli = dict()
        for name in modulus_names:
            if parser[name] is not None:
                moduli[name] = parser[name][0]

        eigenvalues = list(parser['eigenvalues'])[0]

        return matrices, moduli, eigenvalues

    def get_elastic_constants_order3(self):
        path = os.path.join(self.maindir, 'ElaStic_3rd.out')
        elastic_constant = []

        def arrange_matrix(val):
            val = val.strip().split('\n')
            matrix = [v.strip().split() for v in val if v.strip()]
            matrix = np.array(matrix).reshape((12, 18))
            arranged = []
            for i in range(2):
                for j in range(3):
                    arranged.append(
                        matrix[i * 6: (i + 1) * 6, j * 6: (j + 1) * 6].tolist())
            return arranged

        def get_cijk(val):
            val = val.replace('=', '').replace('GPa', '')
            cijk = dict()
            for v in val.split('\n'):
                v = v.split()
                if len(v) == 2:
                    cijk[v[0].strip()] = float(v[1])
            return cijk

        space_group_number = self.info['space_group_number']

        quantities = [
            Quantity(
                'elastic_constant', r'\%\s*\n([\s0-6A-L]*)[\n\s\%1-6\-ij]*([\s0-6A-L]*)\n',
                str_operation=arrange_matrix, dtype=str),
            Quantity(
                'cijk', r'(C\d{3}\s*=\s*[C1-6\s=\n\-\.0-9GPa]*)\n\n', str_operation=get_cijk)
        ]

        parsers = UnstructuredTextFileParser(path, quantities)

        elastic_constant_str = list(parsers['elastic_constant'])[0]

        cijk = list(parsers['cijk'])[0]

        # formulas for the coefficients
        coeff_A = cijk.get('C111', 0) + cijk.get('C112', 0) - cijk.get('C222', 0)
        coeff_B = -(cijk.get('C115', 0) + 3 * cijk.get('C125', 0)) / 2
        coeff_C = (cijk.get('C114', 0) + 3 * cijk.get('C124', 0)) / 2
        coeff_D = -(2 * cijk.get('C111', 0) + cijk.get('C112', 0) - 3 * cijk.get('C222', 0)) / 4
        coeff_E = -cijk.get('C114', 0) - 2 * cijk.get('C124', 0)
        coeff_F = -cijk.get('C115', 0) - 2 * cijk.get('C125', 0)
        coeff_G = -(cijk.get('C115', 0) - cijk.get('C125', 0)) / 2
        coeff_H = (cijk.get('C114', 0) - cijk.get('C124', 0)) / 2
        coeff_I = (2 * cijk.get('C111', 0) - cijk.get('C112', 0) - cijk.get('C222', 0)) / 4
        coeff_J = (cijk.get('C113', 0) - cijk.get('C123', 0)) / 2
        coeff_K = -(cijk.get('C144', 0) - cijk.get('C155', 0)) / 2

        if space_group_number <= 148:  # rhombohedral II
            coefficients = dict(
                A=coeff_A, B=coeff_B, C=coeff_C, D=coeff_D, E=coeff_E, F=coeff_F, G=coeff_G,
                H=coeff_H, I=coeff_I, J=coeff_J, K=coeff_K)
        elif space_group_number <= 167:  # rhombohedral I
            coefficients = dict(
                A=coeff_A, B=coeff_C, C=coeff_D, D=coeff_E, E=coeff_H, F=coeff_I, G=coeff_J,
                H=coeff_K)
        elif space_group_number <= 176:  # hexagonal II
            coefficients = dict(
                A=coeff_A, B=coeff_D, C=coeff_I, D=coeff_J, E=coeff_K)
        elif space_group_number <= 194:  # hexagonal I
            coefficients = dict(
                A=coeff_A, B=coeff_D, C=coeff_I, D=coeff_J, E=coeff_K)
        else:
            coefficients = dict()

        # assign values to the elastic constant matrix from the independent ec and coefficients
        elastic_constant = np.zeros((6, 6, 6))
        for i in range(6):
            for j in range(6):
                for k in range(6):
                    val = elastic_constant_str[i][j][k]
                    if val == '0':
                        elastic_constant[i][j][k] = 0
                    elif val.isdigit():
                        elastic_constant[i][j][k] = cijk['C%s' % val]
                    else:
                        elastic_constant[i][j][k] = coefficients.get(val, 0)

        return elastic_constant



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
All the API flask restplus models.
'''

from typing import Set, List, Dict, Any
from flask_restplus import fields
import datetime
import math
from cachetools import cached

from nomad import config, datamodel, files
from nomad.app.common import RFC3339DateTime
from nomad.datamodel import EntryMetadata
from nomad.datamodel.dft import DFTMetadata
from nomad.datamodel.optimade import OptimadeEntry

from .api import api, url


# TODO error/warning objects

json_api_meta_object_model = api.model('MetaObject', {
    'query': fields.Nested(model=api.model('Query', {
        'representation': fields.String(
            required=True,
            description='A string with the part of the URL following the base URL')
    }, description='Information on the query that was requested.')),

    'api_version': fields.String(
        required=True,
        description='A string containing the version of the API implementation.'),

    'time_stamp': RFC3339DateTime(
        required=True,
        description='A timestamp containing the date and time at which the query was executed.'),

    'data_returned': fields.Integer(
        required=True,
        description='An integer containing the number of data objects returned for the query.'),

    'more_data_available': fields.Boolean(
        required=True,
        description='False if all data for this query has been returned, and true if not.'),

    'provider': fields.Nested(
        required=True, skip_none=True,
        description='Information on the database provider of the implementation.',
        model=api.model('Provider', {
            'name': fields.String(
                required=True,
                description='A short name for the database provider.'),
            'description': fields.String(
                required=True,
                description='A longer description of the database provider.'),
            'prefix': fields.String(
                required=True,
                description='Database-provider-specific prefix.'),
            'homepage': fields.String(
                required=False,
                description='Homepage of the database provider'),
            'index_base_url': fields.String(
                required=False,
                description='Base URL for the index meta-database.')
        })),

    'data_available': fields.Integer(
        required=False,
        description=('An integer containing the total number of data objects available in '
                     'the database.')),

    'last_id': fields.String(
        required=False,
        description='A string containing the last ID returned'),

    'response_message': fields.String(
        required=False,
        description='Response string from the server.'),

    'implementation': fields.Nested(
        required=False, skip_none=True,
        description='Server implementation details.',
        model=api.model('Implementation', {
            'name': fields.String(
                description='Name of the implementation'),
            'version': fields.String(
                description='Version string of the current implementation.'),
            'source_url': fields.String(
                description=' URL of the implementation source, either downloadable archive or version control system.'),
            'maintainer': fields.Nested(
                skip_none=True,
                description='Details about the maintainer of the implementation',
                model=api.model('Maintainer', {
                    'email': fields.String()
                })
            )
        }))
})


class Meta():

    def __init__(
            self, query: str, returned: int, available: int = None, last_id: str = None):

        self.query = dict(representation=query)
        self.api_version = '1.0.0'
        self.time_stamp = datetime.datetime.now()
        self.data_returned = returned
        self.more_data_available = available > returned if available is not None else False
        self.provider = dict(
            name=config.meta.name,
            description=config.meta.description,
            prefix='nomad',
            homepage=config.meta.homepage,
            index_base_url=url(version=None, prefix='index')
        )

        self.data_available = available
        self.last_id = last_id
        self.implementation = dict(
            name='nomad@fairdi',
            version=config.meta.version,
            source_url=config.meta.source_url,
            maintainer=dict(email=config.meta.maintainer_email))


class Links:
    def __init__(self, endpoint: str, returned: int, page_number: int, page_limit: int, **kwargs):
        last_page = math.ceil(returned / page_limit)

        rest = dict(page_limit=page_limit)
        rest.update(**{key: value for key, value in kwargs.items() if value is not None})
        self.self = url()
        self.related = None
        self.first = url(endpoint, page_number=1, **rest)
        self.last = url(endpoint, page_number=last_page, **rest)
        self.prev = url(endpoint, page_number=max((page_number - 1, 1)), **rest)
        self.next = url(endpoint, page_number=min((page_number + 1, last_page)), **rest)


json_api_links_model = api.model('ApiLinks', {
    'base_url': fields.String(
        description='The base URL of the implementation'),

    'next': fields.String(
        description=('A link to fetch the next set of results. When the current response '
                     'is the last page of data, this field MUST be either omitted or null.')),

    'prev': fields.String(
        description=('The previous page of data. null or omitted when the current response '
                     'is the first page of data.')),

    'last': fields.String(
        description='The last page of data.'),

    'first': fields.String(
        description='The first page of data.')
})


json_api_response_model = api.model('Response', {
    'links': fields.Nested(
        required=False,
        description='Links object with pagination links.',
        skip_none=True,
        model=json_api_links_model),

    'meta': fields.Nested(
        required=True, skip_none=True,
        description='JSON API meta object.',
        model=json_api_meta_object_model),

    'included': fields.List(
        fields.Arbitrary(),
        required=False, skip_none=True,
        description=('A list of JSON API resource objects related to the primary data '
                     'contained in data. Responses that contain related resources under '
                     'included are known as compound documents in the JSON API.'))
})

json_api_data_object_model = api.model('DataObject', {
    'type': fields.String(
        description='The type of the object [structure or calculations].'),

    'id': fields.String(
        description='The id of the object.'),

    'attributes': fields.Raw(
        description='A dictionary, containing key-value pairs representing the entries properties'),

    'relationships': fields.Raw(
        description='In accordance with section Relationships, all entry types MAY use relationships to describe relations to other entries.')
})


json_api_info_model = api.model('CalculationInfo', {
    'description': fields.String(
        description='Description of the entry'),

    'properties': fields.Raw(
        description=('A dictionary describing queryable properties for this '
                     'entry type, where each key is a property name')),

    'formats': fields.List(
        fields.String(),
        required=True,
        description='List of output formats available for this type of entry'),

    'output_fields_by_format': fields.Raw(
        description=('Dictionary of available output fields for this entry'
                     'type, where the keys are the values of the formats list'
                     'and the values are the keys of the properties dictionary'))

})


@cached({})
def get_entry_properties(include_optimade: bool = True):
    properties = {
        attr.name: dict(description=attr.description)
        for attr in OptimadeEntry.m_def.all_properties.values()
        if include_optimade}

    def add_nmd_properties(prefix, section_cls):
        for quantity in section_cls.m_def.all_quantities.values():
            name = prefix + quantity.name
            properties[name] = dict(description=quantity.description)

    add_nmd_properties('_nmd_', EntryMetadata)
    add_nmd_properties('_nmd_dft_', DFTMetadata)

    return properties


class EntryDataObject:
    def __init__(self, calc: EntryMetadata, optimade_type: str, response_fields: Set[str] = None):

        def include(key):
            if optimade_type == 'calculations' and key not in ['immutable_id', 'last_modified']:
                return False

            if response_fields is None or (key in response_fields):
                return True

            return False

        attrs = {key: value for key, value in calc.dft.optimade.m_to_dict().items() if include(key)}
        if include('immutable_id'):
            attrs['immutable_id'] = calc.calc_id
        if include('last_modified'):
            attrs['last_modified'] = calc.last_processing if calc.last_processing is not None else calc.upload_time

        if response_fields is not None:
            for request_field in response_fields:
                if not request_field.startswith('_nmd_'):
                    continue

                try:
                    if request_field.startswith('_nmd_dft_'):
                        attrs[request_field] = getattr(calc.dft, request_field[9:])
                    else:
                        attrs[request_field] = getattr(calc, request_field[5:])
                except AttributeError:
                    # if unknown properties where provided, we will ignore them
                    pass

        self.type = optimade_type
        self.id = calc.calc_id
        self.attributes = attrs


class Property:
    @staticmethod
    def from_nomad_to_optimade(name: str):
        if name.startswith('optimade.'):
            return name[9:]
        else:
            return '_nomad_%s' % name

    @staticmethod
    def from_optimade_to_nomad(name: str):
        if name.startswith('_nomad_'):
            return name[7:]
        else:
            return 'optimade.%s' % name


json_api_single_response_model = api.inherit(
    'SingleResponse', json_api_response_model, {
        'data': fields.Nested(
            model=json_api_data_object_model,
            required=True, skip_none=True,
            description=('The returned response object.'))
    })

json_api_list_response_model = api.inherit(
    'ListResponse', json_api_response_model, {
        'data': fields.List(
            fields.Nested(json_api_data_object_model, skip_none=True),
            required=True,
            description=('The list of returned response objects.'))
    })

json_api_info_response_model = api.inherit(
    'InfoResponse', json_api_response_model, {
        'data': fields.Nested(
            model=json_api_info_model,
            required=True,
            description=('The returned response object.'))
    })


base_endpoint_parser = api.parser()
base_endpoint_parser.add_argument(
    'response_format', type=str,
    help=('The output format requested. Defaults to the format string "json", which '
          'specifies the standard output format described in this specification.'))
base_endpoint_parser.add_argument(
    'email_address', type=str,
    help=('An email address of the user making the request. The email SHOULD be that of a '
          'person and not an automatic system.'))
base_endpoint_parser.add_argument(
    'response_fields', action='split', type=str,
    help=('A comma-delimited set of fields to be provided in the output. If provided, only '
          'these fields MUST be returned and no others.'))

entry_listing_endpoint_parser = base_endpoint_parser.copy()
entry_listing_endpoint_parser.add_argument(
    'filter', type=str, help='An optimade filter string.')
entry_listing_endpoint_parser.add_argument(
    'page_limit', type=int,
    help='Sets a numerical limit on the number of entries returned.')
entry_listing_endpoint_parser.add_argument(
    'page_number', type=int,
    help='Sets the page number to return.')
entry_listing_endpoint_parser.add_argument(
    'sort', type=str,
    help='Name of the property to sort the results by.')

single_entry_endpoint_parser = base_endpoint_parser.copy()


def to_calc_with_metadata(results: List[Dict[str, Any]]):
    ''' Translates search results into :class:`EntryMetadata` objects read from archive. '''

    upload_files_cache: Dict[str, files.UploadFiles] = {}

    def transform(result):
        calc_id, upload_id = result['calc_id'], result['upload_id']
        upload_files = upload_files_cache.get(upload_id)

        if upload_files is None:
            upload_files = files.UploadFiles.get(upload_id)
            upload_files_cache[upload_id] = upload_files

        archive = upload_files.read_archive(calc_id)  # , access='public')
        metadata = archive[calc_id]['section_metadata'].to_dict()
        return datamodel.EntryMetadata.m_from_dict(metadata)

    result = [transform(result) for result in results]

    for upload_files in upload_files_cache.values():
        upload_files.close()

    return result



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np


names_plane = {0: 'fix_yz', 1: 'fix_xz', 2: 'fix_xy'}
names_line = {0: 'fix_x', 1: 'fix_y', 2: 'fix_z'}

def get_index(v):
    """ Try to guess the Nomad name
    Parameters:
        v: (3,) arraylike
            vector perpedicular to the plane
            or the direction of the line.
    """
    v /= np.linalg.norm(v)
    v2 = v * v
    perm = v2.argsort()
    d = perm[2]
    return d

def get_nomad_name(c):
    """This tries to find the appropriate named name from 
    the constaints giving by a direction"""
    constraint = c.todict()
    name = str(constraint.get('name'))
    if name == 'FixedPlane':
        d = get_index(c.dir)
        return names_plane[d]
    elif name == 'FixedLine':
       d = get_index(c.dir)
       return names_line[d]
    elif name == 'FixAtoms':
        return 'fix_xyz'
    else:
        return name



from .entries import BaseResourceMapper

__all__ = ("StructureMapper",)


class StructureMapper(BaseResourceMapper):

    ENDPOINT = "structures"

    LENGTH_ALIASES = (
        ("elements", "nelements"),
        ("element_ratios", "nelements"),
        ("cartesian_site_positions", "nsites"),
        ("species_at_sites", "nsites"),
    )



# Copyright 2015-2018 Franz Knuth, Fawzi Mohamed, Wael Chibani, Ankit Kariryaa, Lauri Himanen, Danio Brambila
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
from builtins import object
try:
    import setup_paths
except ImportError:
    pass
import numpy as np
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.unit_conversion.unit_conversion import convert_unit_function
from fhiaimsparser.FhiAimsCommon import get_metaInfo
import logging, os, re, sys

############################################################
# This is the parser for the band.out file of FHI-aims.
############################################################

logger = logging.getLogger("nomad.FhiAimsBandParser")

class FhiAimsBandParserContext(object):
    """Context for parsing FHI-aims band.out file.

    Attributes:
        band_energies: Stores parsed eigenvalues.
        band_k_points Stores parsed k-points.
        band_occupations: Stores parsed occupations.

    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of writing and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self, writeMetaData = True):
        """Args:
            writeMetaData: Deteremines if metadata is written or stored in class attributes.
        """
        self.writeMetaData = writeMetaData

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        # get unit from metadata for band energies
        unit = parser.parserBuilder.metaInfoEnv.infoKinds['band_energies'].units
        # get function for unit conversion, aims energies are in eV
        self.converter = convert_unit_function('eV', unit)
        # allows to reset values if the same superContext is used to parse different files
        self.band_energies = None
        self.band_k_points = None
        self.band_occupations = None

    def onClose_section_k_band(self, backend, gIndex, section):
        """Trigger called when section_k_band is closed.

        Store the parsed values and write them if writeMetaData is True.
        """
        # extract k-points
        band_k = []
        for i in ['1', '2', '3']:
            ki = section['x_fhi_aims_band_k' + i]
            if ki is not None:
                band_k.append(ki)
        if band_k:
            # need to transpose array since its shape is [n_k_points_per_segment,3] in the metadata
            self.band_k_points = np.transpose(np.asarray(band_k))
        # extract occupations and eigenvalues
        band_occupations = []
        band_energies = []
        if section['x_fhi_aims_band_occupations_eigenvalue_string'] is not None:
            for string in section['x_fhi_aims_band_occupations_eigenvalue_string']:
                strings = string.split()
                # first number is occupation and then every second one
                band_occupations.append([float(x) for x in strings[0::2]])
                # second number is eigenvalue and then every second one
                # convert units
                band_energies.append([self.converter(float(x)) for x in strings[1::2]])

        if band_occupations:
            # do not need to transpose array since its shape is [n_k_points,n_eigen_values] in the metadata
            self.band_occupations = np.asarray(band_occupations)
        if band_energies:
            # do not need to transpose array since its shape is [n_k_points,n_eigen_values] in the metadata
            self.band_energies = np.asarray(band_energies)
        # write metadata only if values were found for all three quantities
        if self.writeMetaData:
            if band_k and band_occupations and band_energies:
                # need to add dimension for number_of_k_point_segments
                backend.addArrayValues('band_k_points', np.asarray([self.band_k_points]))
                # need to add dimension for number_of_k_point_segments and number_of_spin_channels
                backend.addArrayValues('band_occupations', np.asarray([[self.band_occupations]]))
                # need to add dimension for number_of_k_point_segments and number_of_spin_channels
                backend.addArrayValues('band_energies', np.asarray([[self.band_energies]]))

def build_FhiAimsBandFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the band.out file of FHI-aims.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses band.out file of FHI-aims.
    """
    return SM (name = 'Root1',
        startReStr = "",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [
        SM (name = 'Root2',
            startReStr = "",
            sections = ['section_single_configuration_calculation'],
            forwardMatch = True,
            weak = True,
            subMatchers = [
            SM (name = 'Root3',
                startReStr = "",
                sections = ['section_k_band'],
                forwardMatch = True,
                weak = True,
                subMatchers = [
                SM (r"\s*[0-9]+\s+(?P<x_fhi_aims_band_k1>[-+0-9.]+)\s+(?P<x_fhi_aims_band_k2>[-+0-9.]+)\s+(?P<x_fhi_aims_band_k3>[-+0-9.]+)\s+(?P<x_fhi_aims_band_occupations_eigenvalue_string>[-+0-9.\s]+)", repeats = True)
                ])
            ])
        ])

def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_k_band': CachingLvl,
                               'section_run': CachingLvl,
                               'section_single_configuration_calculation': CachingLvl,
                              }
    # Set all band metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.startswith('x_fhi_aims_band_'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName

def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the FHI-aims band.out file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get band.out file description
    FhiAimsBandSimpleMatcher = build_FhiAimsBandFileSimpleMatcher()
    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)
    # set parser info
    parserInfo = {'name':'fhi-aims-band-out-parser', 'version': '1.0'}
    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)
    # start parsing
    mainFunction(mainFileDescription = FhiAimsBandSimpleMatcher,
                 metaInfoEnv = metaInfoEnv,
                 parserInfo = parserInfo,
                 cachingLevelForMetaName = cachingLevelForMetaName,
                 superContext = FhiAimsBandParserContext())

if __name__ == "__main__":
    main(CachingLevel.Forward)



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from ase import Atoms
from ase import units
import re


class Reader:
    methods = ['AM1', 'MNDO', 'MNDOD', 'PM3', 'PM6', 'PM6-D3', 'PM6-DH+',
               'PM6-DH2', 'PM6-DH2X', 'PM6-D3H4', 'PM6-D3H4X', 'PMEP', 'PM7',
               'PM7-TS', 'RM1']

    multiplicities = {'singlet': 1, 'doublet': 2, 'triplet': 3, 'quartet': 4,
                      'quintet': 5, 'sextet': 6, 'septet': 7, 'octet': 8,
                      'nonet': 9}

    def __init__(self, filename):

        self.unit2ase = {  # convert to ase units (eV, Angstrom)
            'ev': units.eV,
            'kcal/mol': units.kcal / units.mol,
            'kj/mol': units.kJ / units.mol,
            'hartree': units.Hartree,
            'kcal/angstrom': units.kcal / units.mol / units.Angstrom,
        }

        with open(filename, 'r') as f:
            self.lines = f.readlines()

        self.data = {}
        self.restrs = {
            'program_version':
            re.compile(r"\s*Version\s*(?P<program_version>[0-9.a-zA-Z]+)\s+"),
            'natoms':
            re.compile(r"\s*Empirical\s*Formula:.*=\s*(?P<natoms>[0-9]+)"),
            'positions_begin':
            re.compile(r"\s*CARTESIAN\s*COORDINATES"),
            'energy_total':
            re.compile(r"\s*TOTAL\s*ENERGY\s*=\s*(?P<energy_total>[0-9.+-]+)\s*(?P<unit>[a-zA-Z]+)"),
            'x_mopac_fhof':  # final heat of formation
            re.compile(r"FINAL\s*HEAT\s*OF\s*FORMATION\s*=\s*(?P<x_mopac_fhof>[0-9-.+-]+)\s*(?P<unit>[A-Z/]+)\s*="),
            'forces_begin':
            re.compile(r"\s*FINAL\s*POINT\s*AND\s*DERIVATIVES"),
            'eigenvalues_begin':
            re.compile(r"\s*EIGENVALUES"),
            'n_filled':
            re.compile(r"NO.\s*OF\s*FILLED\s*LEVELS\s*=\s*(?P<n_filled>[0-9]+)"),
            'n_alpha':
            re.compile(r"NO.\s*OF\s*ALPHA\s*ELECTRONS\s*=\s*(?P<n_alpha>[0-9]+)"),
            'n_beta':
            re.compile(r"NO.\s*OF\s*BETA\s*ELECTRONS\s*=\s*(?P<n_beta>[0-9]+)"),
            'spin_S2':
            re.compile(r"\(S\*\*2\)\s*=\s*(?P<spin_S2>[0-9.]+)"),
            'time_calculation':
            re.compile(r"TOTAL\s*JOB\s*TIME:\s*(?P<time_calculation>[0-9.]+)\s*"),
            'total_charge':
            re.compile(r"\s*\**\s*CHARGE\s*ON\s*SYSTEM\s*=\s*(?P<total_charge>[0-9]+)")

        }
        self.read()
        self.calculate_and_transform()

    def rep2bl(self, j, fun=float):
        """ Loop until a blank line is found, starting from line j.
            Parameters:
                j: starting line
                fun: [fun(x) for x in line.split()]
        """
        x = []
        while not self.lines[j].isspace():  # continue until a blank line
            x += [fun(e) for e in self.lines[j].split()]
            j += 1
        return x

    def get_index(self, pattern, istart=0, istop=None):
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        lines = self.lines[istart:istop]
        for i, line in enumerate(lines):
            m = re.search(pattern, line)
            if m:
                break
        return istart + i

    def get_input_parameter_line(self):
        i = self.get_index(r"\s*\*\s*CALCULATION DONE:")
        j = self.get_index(r"\s*\*{3,}", istart=i)
        return self.lines[j + 1]

    def get_natoms(self):
        natoms = None
        for line in self.lines:
            m = re.search(self.restrs['natoms'],
                          line)
            if m:
                natoms = int(m.group('natoms'))
                break
        assert natoms
        return natoms

    def read(self):
        natoms = self.get_natoms()
        self.data['natoms'] = natoms
        self.data['inp_parm_line'] = self.get_input_parameter_line().rstrip()

        for i, line in enumerate(self.lines):
            # program_version
            if self.data.get('program_version') is None:
                m = re.search(self.restrs['program_version'], line)
                if m:
                    self.data['program_version'] = m.group('program_version')

            if self.data.get('total_charge') is None:
                m = re.search(self.restrs['total_charge'], line)
                if m:
                    self.data['total_charge'] = int(m.group('total_charge'))

            # atom_positions, atom_labels
            if self.data.get('atom_positions') is None:
                m = re.search(self.restrs['positions_begin'], line)
                if m:
                    if 'ATOM' not in self.lines[i + 2]:
                        begin = 2
                        symbols = []
                        positions = []
                        for l in self.lines[i + begin: i + begin + natoms]:
                            vals = l.split()
                            symbols.append(vals[1])
                            positions.append([float(val) for val in vals[2:5]])

                        self.data['atom_labels'] = np.asarray(symbols)
                        self.data['atom_positions'] = np.asarray(positions)

            # total_energy
            if self.data.get('energy_total') is None:
                m = re.search(self.restrs['energy_total'], line)
                if m:
                    self.data['energy_total'] = (
                        float(m.group('energy_total')) *
                        self.unit2ase[m.group('unit').lower()])

            if self.data.get('x_mopac_fhof') is None:
                m = re.search(self.restrs['x_mopac_fhof'], line)
                if m:
                    self.data['x_mopac_fhof'] = (
                        float(m.group('x_mopac_fhof')) *
                        self.unit2ase[m.group('unit').lower()])
            # forces
            if self.data.get('atom_forces') is None:
                m = re.search(self.restrs['forces_begin'], line)
                if m:
                    forces = []
                    for sline in self.lines[i + 3: i + 3 + natoms * 3]:
                        tmp = sline.split()
                        forces += [-float(tmp[6]) *
                                   self.unit2ase[tmp[7].lower()]]
                    forces = np.array(forces).reshape(natoms, 3)
                    self.data['atom_forces'] = forces

            # eigenvalues
            if self.data.get('n_filled') is None:
                m = re.search(self.restrs['n_filled'], line)
                if m:
                    self.data['n_filled'] = int(m.group('n_filled'))

            if self.data.get('n_alpha') is None:
                m = re.search(self.restrs['n_alpha'], line)
                if m:
                    self.data['n_alpha'] = int(m.group('n_alpha'))

            if self.data.get('n_beta') is None:
                m = re.search(self.restrs['n_beta'], line)
                if m:
                    self.data['n_beta'] = int(m.group('n_beta'))

            if self.data.get('eigenvalues_values') is None:
                m = re.search(self.restrs['eigenvalues_begin'], line)
                if m:
                    if 'ALPHA' in line:  # spin polarized calculation
                        eigs_a = self.rep2bl(i + 1, float)
                    elif 'BETA' in line:
                        eigs_b = self.rep2bl(i + 1, float)
                        eigs = np.array([eigs_a, eigs_b]).reshape(2, 1, -1)
                        self.data['eigenvalues_values'] = eigs
                    else:
                        eigs = np.array(self.rep2bl(i + 1,
                                                    float)).reshape(1, 1, -1)
                        self.data['eigenvalues_values'] = eigs

            if self.data.get('spin_S2') is None:
                m = re.search(self.restrs['spin_S2'], line)
                if m:
                    self.data['spin_S2'] = float(m.group('spin_S2'))

            if self.data.get('time_calculation') is None:
                m = re.search(self.restrs['time_calculation'], line)
                if m:
                    self.data['time_calculation'] = float(
                        m.group('time_calculation'))

        self.atoms = Atoms(self.data['atom_labels'],
                           positions=self.data['atom_positions'])

    def calculate_and_transform(self):
        # setup occupations, and homo, lumo, somo
        if self.data.get('total_charge') is None:
            self.data['total_charge'] = 0
        eigs = self.data.get('eigenvalues_values')
        if eigs is not None:
            occs = np.zeros(eigs.shape, float)
            nspin = eigs.shape[0]
            self.data['eigenvalues_kpoints'] = np.zeros((1, 3))  # Gamma only
            if nspin == 2:
                n_a = self.data.get('n_alpha')
                n_b = self.data.get('n_beta')
                self.data['number_of_electrons'] = np.array((n_a, n_b), float)
                occs[0, 0, :n_a] = 1.0
                occs[1, 0, :n_b] = 1.0
                self.data['energy_reference_highest_occupied'] = np.array(
                    (eigs[0, 0, n_a - 1], eigs[1, 0, n_b - 1]))
            else:
                n_f = self.data.get('n_filled')
                self.data['number_of_electrons'] = np.array((2 * n_f, ), float)
                occs[0, 0, :n_f] = 1.0
                self.data['energy_reference_highest_occupied'] = np.array(
                    (eigs[0, 0, n_f - 1],))

            self.data['eigenvalues_occupation'] = occs

        if self.data.get('inp_parm_line') is not None:
            inp_parm_line = self.data.get('inp_parm_line')
            if '1SCF' in inp_parm_line:
                structure_optimization = False
            else:
                structure_optimization = True
            self.data['x_mopac_optimization'] = structure_optimization

            for multiplicity, value in self.multiplicities.items():
                if multiplicity in [p.lower() for p in inp_parm_line.split()]:
                    self.data['spin_target_multiplicity'] = value
                    break

            for method in self.methods:
                if method in inp_parm_line:
                    self.data['x_mopac_method'] = method
                    break

    def __getattr__(self, name):
        return self.data.get(name)

    def __contains__(self, name):
        return name in self.data

if __name__ == '__main__':
    import sys
    fname = sys.argv[1]
    r = Reader(fname)
    print(r.eigenvalues_values.shape)
#    i = r.get_index(r"CALCULATION DONE:")
#    print('line number {0}| '.format(i) + r.lines[i])
    print('printing r.data')
    for key, val in r.data.items():
        if isinstance(val, np.ndarray):
            print(key, type(val), '| shape:', val.shape)
            if len(val.flat) <= 40:
                print(val)
        else:
            print(key, val)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition


m_package = Package(
    name='general_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='general.nomadmetainfo.json'))


class section_entry_info(MSection):
    '''
    General information about this entry that is independent from its domain, field, or
    used parser
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_entry_info'))

    entry_upload_time = Quantity(
        type=np.dtype(np.int64),
        shape=[],
        description='''
        Upload datetime, given as total number of seconds is the elapsed since the unix
        epoch (1 January 1970)
        ''',
        a_legacy=LegacyDefinition(name='entry_upload_time'))

    entry_uploader_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the uploader, given as lastname, firstname.
        ''',
        a_legacy=LegacyDefinition(name='entry_uploader_name'))

    entry_uploader_id = Quantity(
        type=str,
        shape=[],
        description='''
        The id of the uploader.
        ''',
        a_legacy=LegacyDefinition(name='entry_uploader_id'))

    upload_id = Quantity(
        type=str,
        shape=[],
        description='''
        Nomad upload id
        ''',
        a_legacy=LegacyDefinition(name='upload_id'))

    calc_id = Quantity(
        type=str,
        shape=[],
        description='''
        Nomad calc id.
        ''',
        a_legacy=LegacyDefinition(name='calc_id'))

    calc_hash = Quantity(
        type=str,
        shape=[],
        description='''
        Calculation hash based on raw file contents.
        ''',
        a_legacy=LegacyDefinition(name='calc_hash'))

    mainfile = Quantity(
        type=str,
        shape=[],
        description='''
        Path to the main file within the upload.
        ''',
        a_legacy=LegacyDefinition(name='mainfile'))

    parser_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the parser used to extract this information.
        ''',
        a_legacy=LegacyDefinition(name='parser_name'))

    filepaths = Quantity(
        type=str,
        shape=['number_of_files'],
        description='''
        Filepaths of files that belong to this entry, i.e. files in the same directory.
        Filepaths are relative to the upload.
        ''',
        a_legacy=LegacyDefinition(name='filepaths'))

    number_of_files = Quantity(
        type=int,
        shape=[],
        description='''
        Number of files that belong to this entry.
        ''',
        a_legacy=LegacyDefinition(name='number_of_files'))

    section_archive_processing_info = SubSection(
        sub_section=SectionProxy('section_archive_processing_info'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_archive_processing_info'))


class section_archive_processing_info(MSection):
    '''
    Information about the used archive processing steps and their execution.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_archive_processing_info'))

    archive_processor_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the applied archive processing program.
        ''',
        a_legacy=LegacyDefinition(name='archive_processor_name'))

    archive_processor_error = Quantity(
        type=str,
        shape=[],
        description='''
        The main error during execution of the archive processing program that failed the
        program.
        ''',
        a_legacy=LegacyDefinition(name='archive_processor_error'))

    number_of_archive_processor_warnings = Quantity(
        type=int,
        shape=[],
        description='''
        Number of warnings during execution of the archive processing program.
        ''',
        a_legacy=LegacyDefinition(name='number_of_archive_processor_warnings'))

    archive_processor_warnings = Quantity(
        type=str,
        shape=['number_of_archive_processor_warnings'],
        description='''
        Warnings during execution of the archive processing program.
        ''',
        a_legacy=LegacyDefinition(name='archive_processor_warnings'))

    archive_processor_status = Quantity(
        type=str,
        shape=[],
        description='''
        Status returned by archive processing program.
        ''',
        a_legacy=LegacyDefinition(name='archive_processor_status'))


m_package.__init_metainfo__()
